# -*- coding: utf-8 -*-
"""
This module provides an interface into the transformers BERT models for fine tuning.
"""
import warnings
import os

import transformers
import torch
import torch.nn.functional as F

from mangoes.modeling.training_utils import MangoesTextClassificationDataset, MangoesQuestionAnsweringDataset,\
    CoreferenceFineTuneTrainer, MangoesCoreferenceDataset, MangoesMultipleChoiceDataset, \
    MultipleLearnRateFineTuneTrainer, freeze_base_layers
from mangoes.modeling.bert_base import PipelineMixin, TransformerModel
from mangoes.modeling.coref import BertForCoreferenceResolutionBase


class BERTForSequenceClassification(PipelineMixin, TransformerModel):
    """
    BERT model with Sequence classification head (linear layer on top of pooled output)

    :var self.tokenizer: transformers.BertTokenizerFast object,
        see https://huggingface.co/transformers/model_doc/bert.html#berttokenizerfast for more documentation
    :var self.model: transformers.BertForSequenceClassification object,
        see https://huggingface.co/transformers/model_doc/bert.html#bertforsequenceclassification

    The init function should rarely be used to directly instantiate this object, as it will initialize an untrained
    base BERT model; use the :func:`~load` function instead, as this can be used to load a pretrained base BERT
    model for fine-tuning (or a pretrained base and fine-tuning head(s) for inference or further fine-tuning).
    This follows the transformers library framework for saving/loading.

    Parameters
    ----------
    pretrained_tokenizer: str
        Either:
            - A string with the `shortcut name` of a pretrained tokenizer to load from cache or download, e.g.,
              ``bert-base-uncased``.
            - A string with the `identifier name` of a pretrained tokenizer that was user-uploaded to our S3, e.g.,
              ``dbmdz/bert-base-german-cased``.
            - A path to a `directory` containing tokenizer vocab or file saved using
              :func:`tokenizer.save_model` or :func:`tokenizer.save`, e.g., ``./my_model_directory/``.
    labels: list of str
        List of class names. Will raise error if labels=None and label2id=None
    label2id: dict of str -> int
        dict mapping class name to index of output layer
        if None, will create from labels.
    device: int, optional, defaults to None
        if -1, use cpu, if >= 0, use CUDA device number. If None, will use GPU if available
    **keyword_args include arguments passed to transformers.BertConfig (see mangoes.modeling.BERTBase docstring)
    """
    def __init__(self, pretrained_tokenizer, labels=None, label2id=None, device=None, **keyword_args):
        TransformerModel.__init__(self, device)
        PipelineMixin.__init__(self)
        if label2id is None and labels is None:
            raise RuntimeError("Must provide either labels or label to id mapping")
        if label2id is None:
            self.label2id = {tag: id for id, tag in enumerate(set(labels))}
        else:
            self.label2id = label2id
        self.id2label = {id: tag for tag, id in self.label2id.items()}
        self.tokenizer = transformers.BertTokenizerFast.from_pretrained(pretrained_tokenizer)
        keyword_args["vocab_size"] = len(self.tokenizer.get_vocab())
        config = transformers.BertConfig(id2label=self.id2label, **keyword_args)
        self.model = transformers.BertForSequenceClassification(config)

    def _construct_pipeline(self):
        """
        Implementation for creating inference pipeline.
        """
        self.model.eval()
        self.pipeline = transformers.TextClassificationPipeline(model=self.model, tokenizer=self.tokenizer,
                                                                device=-1 if self.model.device.type == "cpu" else
                                                                self.model.device.index)

    def predict(self, inputs, return_all_scores=False):
        """
        Predicts classes for input texts.

        Parameters
        ----------
        inputs: str or list of strs
            inputs to classify
        return_all_scores: Boolean (default=False)
            Whether to return all scores or just the predicted class score

        Returns
        -------
        list of dict, or list of list of dict if return_all_scores=True
        If one sequence is passed as input, a list with one element (either a dict or list of dicts if
            return_all_scores=True) will be returned.
        for each input, a dict containing:
            label (str) – The label predicted.
            score (float) – The corresponding probability.
        if return_all_scores, dict will be returned for each class, for each input
        """
        return self._predict(inputs, return_all_scores=return_all_scores)

    @classmethod
    def load(cls, pretrained_tokenizer, pretrained_model, labels=None, label2id=None, device=None):
        """
        Load a mangoes BERTForSequenceClassification object from saved tokenizer and model files.
        This is the preferred way to initialize this class, as the base BERT model should be pretrained.
        This function follows the transformers library way of loading pretrained models for fine-tuning, which allows
        for the following use-cases:
            - Load just the base BERT model, for use in fine-tuning.
            - Load the base BERT model and the fine tuned heads, for use in inference or further fine-tuning.

        Parameters
        ----------
        pretrained_tokenizer: str
            Either:
                - A string with the `shortcut name` of a pretrained model to load from cache or download, e.g.,
                  ``bert-base-uncased``.
                - A string with the `identifier name` of a pretrained model that was user-uploaded to our S3, e.g.,
                  ``dbmdz/bert-base-german-cased``.
                - A path to a `directory` containing model weights saved using
                  :func:`~save_pretrained`, e.g., ``./my_model_directory/``.
        pretrained_model: str
            Either:
                - A string with the `shortcut name` of a pretrained model to load from cache or download, e.g.,
                  ``bert-base-uncased``.
                - A string with the `identifier name` of a pretrained model that was user-uploaded to our S3, e.g.,
                  ``dbmdz/bert-base-german-cased``.
                - A path to a `directory` containing model weights saved using
                  :func:`~save_pretrained`, e.g., ``./my_model_directory/``.
        labels: list of str
            List of class names. Will raise error if labels=None and label2id=None
        label2id: dict of str -> int
            dict mapping class name to index of output layer
            if None, will create from labels.
        device: int, optional, defaults to None
            if -1, use cpu, if >= 0, use CUDA device number. If None, will use GPU if available
        """
        if label2id is None and labels is None:
            raise RuntimeError("Must provide either labels or label to id mapping")
        elif label2id is None:
            label2id = {tag: id for id, tag in enumerate(set(labels))}
        id2label = {id: tag for tag, id in label2id.items()}
        model = transformers.BertForSequenceClassification.from_pretrained(pretrained_model, id2label=id2label)
        model_object = cls(pretrained_tokenizer, labels=labels, label2id=label2id, device=device)
        model_object.model = model
        model_object.model.eval()
        return model_object

    def train(self, output_dir=None, train_text=None, train_targets=None, eval_text=None, eval_targets=None,
              max_len=None, freeze_base=False, task_learn_rate=None, collator=None, train_dataset=None,
              eval_dataset=None, compute_metrics=None, trainer=None, **training_args):
        """
        Fine tune a BERT model on a text classification dataset

        Parameters
        ----------
        output_dir: str
            Path to the output directory where the model predictions and checkpoints will be written. Used to
            instantiate Trainer if trainer argument is None.
        train_text: List[str]
            list of training texts
        train_targets: List[int] or List [str]
            corresponding list of classes for each training text. If strings, will use label2id to convert to output
            indices, else will assume already converted.
        eval_text: (Optional) str or List[str]
            list of evaluation texts
        eval_targets: List[int]
            corresponding list of classes for each evaluation text
        max_len: int
            max length of input sequence. Will default to self.tokenizer.max_length() if None
        freeze_base: Boolean
            Whether to freeze the weights of the base BERT model, so training only changes the task head weights.
            If true, the requires_grad flag for parameters of the base model will be set to false before training.
        task_learn_rate: float
            Learning rate to be used for task specific parameters, (base parameters will use the normal, ie already
            defined in **training_args, learning rate). If None, all parameters will use the same normal learning rate.
        collator: Transformers.DataCollator
            custom collator to use
        train_dataset, eval_dataset: torch.Dataset
            instantiated custom dataset object
        compute_metrics: function
            The function that will be used to compute metrics at evaluation. Must return a dictionary string to metric
            values. Used by the trainer, see https://huggingface.co/transformers/training.html#trainer for more info.
        trainer: Transformers.Trainer
            custom instantiated trainer to use
        training_args:
            keyword arguments for training. For complete list, see
            https://huggingface.co/transformers/main_classes/trainer.html#transformers.TrainingArguments
        """
        if freeze_base:
            freeze_base_layers(self.model)
        if not trainer:
            if not output_dir:
                raise RuntimeError("Must provide output directory argument to train() method if trainer argument is "
                                   "None")
            if not train_dataset:
                if not train_text or not train_targets:
                    raise RuntimeError("Incomplete training data provided to train method")
                train_dataset = MangoesTextClassificationDataset(train_text, train_targets, self.tokenizer,
                                                                 max_len=max_len, label2id=self.label2id)
            if eval_text and eval_targets and not eval_dataset:
                eval_dataset = MangoesTextClassificationDataset(eval_text, eval_targets, self.tokenizer,
                                                                max_len=max_len, label2id=self.label2id)
            elif (eval_text or eval_targets) and not eval_dataset:
                warnings.warn("Incomplete evaluation dataset provided to train(). Please include both texts and "
                              "targets. Skipping evaluation dataset.")
            if eval_dataset is not None and "evaluation_strategy" not in training_args:
                training_args["evaluation_strategy"] = "epoch"
            train_args = transformers.TrainingArguments(output_dir=output_dir, **training_args)
            if task_learn_rate is not None:
                trainer = MultipleLearnRateFineTuneTrainer(task_learn_rate=task_learn_rate, model=self.model,
                                                           args=train_args, train_dataset=train_dataset,
                                                           data_collator=collator, eval_dataset=eval_dataset,
                                                           tokenizer=self.tokenizer, compute_metrics=compute_metrics)
            else:
                trainer = transformers.Trainer(self.model, args=train_args, train_dataset=train_dataset,
                                               data_collator=collator, eval_dataset=eval_dataset,
                                               tokenizer=self.tokenizer, compute_metrics=compute_metrics)
        trainer.train()
        self.model.eval()


class BERTForTokenClassification(PipelineMixin, TransformerModel):
    """
    BERT model with Sequence classification head (linear layer on top of pooled output)

    :var self.tokenizer: transformers.BertTokenizerFast object,
        see https://huggingface.co/transformers/model_doc/bert.html#berttokenizerfast for more documentation
    :var self.model: transformers.BertForTokenClassification object,
        see https://huggingface.co/transformers/model_doc/bert.html#bertfortokenclassification

    The init function should rarely be used to directly instantiate this object, as it will initialize an untrained
    base BERT model; use the :func:`~load` function instead, as this can be used to load a pretrained base BERT
    model for fine-tuning (or a pretrained base and fine-tuning head(s) for inference or further fine-tuning).
    This follows the transformers library framework for saving/loading.

    Parameters
    ----------
    pretrained_tokenizer: str
        Either:
            - A string with the `shortcut name` of a pretrained tokenizer to load from cache or download, e.g.,
              ``bert-base-uncased``.
            - A string with the `identifier name` of a pretrained tokenizer that was user-uploaded to our S3, e.g.,
              ``dbmdz/bert-base-german-cased``.
            - A path to a `directory` containing tokenizer vocab or file saved using
              :func:`tokenizer.save_model` or :func:`tokenizer.save`, e.g., ``./my_model_directory/``.
    labels: list of str
        list of class names. Will raise error if labels=None and label2id=None.
    label2id: dict of str -> int
        dict mapping class name to index of output layer
        if None, will create from labels.
    device: int, optional, defaults to None
        if -1, use cpu, if >= 0, use CUDA device number. If None, will use GPU if available
    **keyword_args include arguments passed to transformers.BertConfig (see mangoes.modeling.BERTBase docstring)
    """
    def __init__(self, pretrained_tokenizer, labels=None, label2id=None, device=None, **keyword_args):
        TransformerModel.__init__(self, device)
        PipelineMixin.__init__(self)
        if label2id is None and labels is None:
            raise RuntimeError("Must provide either labels or label to id mapping")
        if label2id is None:
            self.label2id = {tag: id for id, tag in enumerate(set(labels))}
        else:
            self.label2id = label2id
        self.id2label = {id: tag for tag, id in self.label2id.items()}
        self.tokenizer = transformers.BertTokenizerFast.from_pretrained(pretrained_tokenizer)
        keyword_args["vocab_size"] = len(self.tokenizer.get_vocab())
        config = transformers.BertConfig(id2label=self.id2label, **keyword_args)
        self.model = transformers.BertForTokenClassification(config)

    def _construct_pipeline(self):
        """
        Implementation for creating inference pipeline.
        """
        self.model.eval()
        self.pipeline = transformers.TokenClassificationPipeline(model=self.model, tokenizer=self.tokenizer,
                                                                 device=-1 if self.model.device.type == "cpu" else
                                                                 self.model.device.index)

    def predict(self, inputs):
        """
        Predicts classes for input texts

        Parameters
        ----------
        inputs: str or list of strs
            inputs to classify

        Returns
        -------
        list of list of dict. For each sequence, a list of token prediction dicts. If a single sequence is passed as
        input, the output will be a list of 1 list of dictionaries.
        for each token, a dict containing:
            word (str) – The token/word classified.
            score (float) – The corresponding probability for entity.
            entity (str) – The entity predicted for that token/word.
            index (int, only present when self.grouped_entities=False) – The index of the corresponding token in the
                sentence.
        """
        prediction = self._predict(inputs)
        if isinstance(inputs, str):
            return [prediction]
        return prediction

    @classmethod
    def load(cls, pretrained_tokenizer, pretrained_model, labels=None, label2id=None, device=None):
        """
        Load a mangoes BERTForTokenClassification object from saved tokenizer and model files.
        This is the preferred way to initialize this class, as the base BERT model should be pretrained.
        This function follows the transformers library way of loading pretrained models for fine-tuning, which allows
        for the following use-cases:
            - Load just the base BERT model, for use in fine-tuning.
            - Load the base BERT model and the fine tuned heads, for use in inference or further fine-tuning.

        Parameters
        ----------
        pretrained_tokenizer: str
            Either:
                - A string with the `shortcut name` of a pretrained model to load from cache or download, e.g.,
                  ``bert-base-uncased``.
                - A string with the `identifier name` of a pretrained model that was user-uploaded to our S3, e.g.,
                  ``dbmdz/bert-base-german-cased``.
                - A path to a `directory` containing model weights saved using
                  :func:`~save_pretrained`, e.g., ``./my_model_directory/``.
        pretrained_model: str
            Either:
                - A string with the `shortcut name` of a pretrained model to load from cache or download, e.g.,
                  ``bert-base-uncased``.
                - A string with the `identifier name` of a pretrained model that was user-uploaded to our S3, e.g.,
                  ``dbmdz/bert-base-german-cased``.
                - A path to a `directory` containing model weights saved using
                  :func:`~save_pretrained`, e.g., ``./my_model_directory/``.
        labels: list of str
            List of class names. Will raise error if labels=None and label2id=None.
        label2id: dict of str -> int
            dict mapping class name to index of output layer
            if None, will create from labels.
        device: int, optional, defaults to None
            if -1, use cpu, if >= 0, use CUDA device number. If None, will use GPU if available
        """
        if label2id is None and labels is None:
            raise RuntimeError("Must provide either labels or label to id mapping")
        if label2id is None:
            label2id = {tag: id for id, tag in enumerate(set(labels))}
        id2label = {id: tag for tag, id in label2id.items()}
        model = transformers.BertForTokenClassification.from_pretrained(pretrained_model, id2label=id2label)
        model_object = cls(pretrained_tokenizer, labels=labels, label2id=label2id, device=device)
        model_object.model = model
        model_object.model.eval()
        return model_object

    def train(self, output_dir=None, train_text=None, train_targets=None, eval_text=None, eval_targets=None,
              max_len=None, freeze_base=False, task_learn_rate=None, collator=None, train_dataset=None,
              eval_dataset=None, compute_metrics=None, trainer=None, **training_args):
        """
        Fine tune a BERT model on a text classification dataset

        Parameters
        ----------
        output_dir: str
            Path to the output directory where the model predictions and checkpoints will be written. Used to
            instantiate Trainer if trainer argument is None.
        train_text: List[str]
            list of training texts
        train_targets: List[List[int]]
            corresponding list of classes for each token in each training text
        eval_text: str or List[str]
            list of evaluation texts
        eval_targets: List[int]
            corresponding list of classes for each evaluation text
        max_len: int
            max length of input sequence. Will default to self.tokenizer.max_length() if None
        freeze_base: Boolean
            Whether to freeze the weights of the base BERT model, so training only changes the task head weights.
            If true, the requires_grad flag for parameters of the base model will be set to false before training.
        task_learn_rate: float
            Learning rate to be used for task specific parameters, (base parameters will use the normal, ie already
            defined in **training_args, learning rate). If None, all parameters will use the same normal learning rate.
        collator: Transformers.DataCollator
            custom collator to use
        train_dataset, eval_dataset: torch.Dataset
            instantiated custom dataset object
        compute_metrics: function
            The function that will be used to compute metrics at evaluation. Must return a dictionary string to metric
            values. Used by the trainer, see https://huggingface.co/transformers/training.html#trainer for more info.
        trainer: Transformers.Trainer
            custom instantiated trainer to use
        training_args:
            keyword arguments for training. For complete list, see
            https://huggingface.co/transformers/main_classes/trainer.html#transformers.TrainingArguments
        """
        if freeze_base:
            freeze_base_layers(self.model)
        # TODO: allow token classification of tokens outside of pretrained tokenizer's vocabulary.
        # see https://huggingface.co/transformers/custom_datasets.html#token-classification-with-w-nut-emerging-entities
        if not trainer:
            if not output_dir:
                raise RuntimeError("Must provide output directory argument to train() method if trainer argument is "
                                   "None")
            if not train_dataset:
                if not train_text or not train_targets:
                    raise RuntimeError("Incomplete training data provided to train method")
                train_dataset = MangoesTextClassificationDataset(train_text, train_targets, self.tokenizer,
                                                                 max_len=max_len, label2id=self.label2id)
            if eval_text and eval_targets and not eval_dataset:
                eval_dataset = MangoesTextClassificationDataset(eval_text, eval_targets, self.tokenizer,
                                                                max_len=max_len, label2id=self.label2id)
            elif (eval_text or eval_targets) and not eval_dataset:
                warnings.warn("Incomplete evaluation dataset provided to train(). Please include both texts and "
                              "targets. Skipping evaluation dataset.")
            if eval_dataset is not None and "evaluation_strategy" not in training_args:
                training_args["evaluation_strategy"] = "epoch"
            train_args = transformers.TrainingArguments(output_dir=output_dir, **training_args)
            if task_learn_rate is not None:
                trainer = MultipleLearnRateFineTuneTrainer(task_learn_rate=task_learn_rate, model=self.model,
                                                           args=train_args, train_dataset=train_dataset,
                                                           data_collator=collator, eval_dataset=eval_dataset,
                                                           tokenizer=self.tokenizer, compute_metrics=compute_metrics)
            else:
                trainer = transformers.Trainer(self.model, args=train_args, train_dataset=train_dataset,
                                               data_collator=collator, eval_dataset=eval_dataset,
                                               tokenizer=self.tokenizer, compute_metrics=compute_metrics)
        trainer.train()
        self.model.eval()


class BERTForQuestionAnswering(PipelineMixin, TransformerModel):
    """
    BERT model with span classification head on top for extractive question-answering tasks like SQuAD (a linear layer
    on top of the hidden-states output to compute span start logits and span end logits)

    :var self.tokenizer: transformers.BertTokenizerFast object,
        see https://huggingface.co/transformers/model_doc/bert.html#berttokenizerfast for more documentation
    :var self.model: transformers.BertForQuestionAnswering object,
        see https://huggingface.co/transformers/model_doc/bert.html#bertforquestionanswering

    The init function should rarely be used to directly instantiate this object, as it will initialize an untrained
    base BERT model; use the :func:`~load` function instead, as this can be used to load a pretrained base BERT
    model for fine-tuning (or a pretrained base and fine-tuning head(s) for inference or further fine-tuning).
    This follows the transformers library framework for saving/loading.

    Parameters
    ----------
    pretrained_tokenizer: str
        Either:
            - A string with the `shortcut name` of a pretrained tokenizer to load from cache or download, e.g.,
              ``bert-base-uncased``.
            - A string with the `identifier name` of a pretrained tokenizer that was user-uploaded to our S3, e.g.,
              ``dbmdz/bert-base-german-cased``.
            - A path to a `directory` containing tokenizer vocab or file saved using
              :func:`tokenizer.save_model` or :func:`tokenizer.save`, e.g., ``./my_model_directory/``.
    device: int, optional, defaults to None
        if -1, use cpu, if >= 0, use CUDA device number. If None, will use GPU if available
    **keyword_args include arguments passed to transformers.BertConfig (see mangoes.modeling.BERTBase docstring)
    """
    def __init__(self, pretrained_tokenizer, device=None, **keyword_args):
        TransformerModel.__init__(self, device)
        PipelineMixin.__init__(self)
        self.tokenizer = transformers.BertTokenizerFast.from_pretrained(pretrained_tokenizer)
        keyword_args["vocab_size"] = len(self.tokenizer.get_vocab())
        config = transformers.BertConfig(**keyword_args)
        self.model = transformers.BertForQuestionAnswering(config)

    def generate_outputs(self, question, context, pre_tokenized=False, output_attentions=False,
                         output_hidden_states=False, word_embeddings=False):
        """
        Tokenize questions and context and pass them through the BERT model and QA head,
        optionally outputting hidden states or attention matrices.
        Works for single question/context or batch. If a single question/context is given, a batch of size 1 will be
        created.

        Parameters
        ----------
        question: str or List[str] if pre_tokenized=False, else List[str] or List[List[str]]
            The question text
        context: str or List[str] if pre_tokenized=False, else List[str] or List[List[str]]
            The context text
        pre_tokenized: Boolean
            Whether or not the input text is pretokenized (ie, split on spaces)
        output_attentions: Boolean, optional, defaults to False
            Whether or not to return the attentions tensors of all attention layers.
        output_hidden_states: Boolean, optional, defaults to False
            Whether or not to return the hidden states of all layers.
        word_embeddings: Boolean
            Whether or not to filter special token embeddings and average subword embeddings (hidden states) into word
            embeddings. This functionality is not available for this task class. Use the feature extraction class
            instead.

        Returns
        -------
        Dict containing (note that if single text sequence is pass as input, batch size will be 1):
        start_logits: torch.FloatTensor of shape (batch_size, sequence_length)
            Span-start scores (before SoftMax).
        end_logits: torch.FloatTensor of shape (batch_size, sequence_length)
            Span-end scores (before SoftMax).
        offset_mappings: Tensor of shape (batch_size, sequence_length, 2)
            Tensor containing (char_start, char_end) for each token, giving index into input strings of start and end
            character for each token. If input is pre-tokenized, start and end index maps into associated word. Note
            that special tokens are included with 0 for start and end indices, as these don't map into input text
            because they are added inside the function. Offset mappings for both questions and answers are merged to one
            row, but indices still align to them separately.
        hidden_states: Tuple (one for each layer) of torch.FloatTensor (batch_size, sequence_length, hidden_size).
            Hidden-states of the model at the output of each layer plus the initial embedding outputs.
            Only returned if output_hidden_states is True
        attentions: Tuple of torch.FloatTensor (one for each layer) of shape (batch_size, num_heads, sequence_length,
            sequence_length).
            Attentions weights after the attention softmax, used to compute the weighted average in the self-attention
            heads.
            Only returned if output_attentions is True
        """
        question_first = bool(self.tokenizer.padding_side == "right")
        if (isinstance(question, str) or (isinstance(question[0], str) and pre_tokenized)) and \
                (isinstance(context, str) or (isinstance(context[0], str) and pre_tokenized)):
            # if single question and context
            question = [question]
            context = [context]
        elif isinstance(question, list) and (isinstance(context, str) or (isinstance(context[0], str) and
                                                                          pre_tokenized)):
            # if one context for multiple questions
            context = [context for _ in range(len(question))]
        elif isinstance(question, list) and isinstance(context, list) and len(question) != len(context):
            raise RuntimeError("Questions and contexts don't have the same lengths")

        inputs = self.tokenizer(
            text=question if question_first else context,
            text_pair=context if question_first else question,
            padding=True,
            truncation="only_second" if question_first else "only_first",
            is_split_into_words=pre_tokenized,
            max_length=384,
            stride=128,
            return_tensors="pt",
            return_token_type_ids=True,
            return_offsets_mapping=True,
        )
        offset_mappings = inputs.pop('offset_mapping')
        self.model.eval()
        if word_embeddings:
            warnings.warn("Word embedding consolidation not available for Question Answering model. Consider using the"
                          "mangoes.modeling.BERTBase class for feature extractions")
        with torch.no_grad():
            if self.model_device.type == "cuda":
                inputs = {name: tensor.to(self.model_device) for name, tensor in inputs.items()}
                if not next(self.model.parameters()).is_cuda:
                    self.model.to(self.model_device)
            results = self.model.forward(**inputs, return_dict=True,
                                         output_attentions=output_attentions,
                                         output_hidden_states=output_hidden_states)
        results["offset_mappings"] = offset_mappings
        return dict(results)

    def _construct_pipeline(self):
        """
        Implementation for creating inference pipeline.
        """
        self.model.eval()
        # TODO: (02/12/2020) currently the FastTokenizers don't work with QAPiplines, so we have to typecast to slow tok
        # https://github.com/huggingface/transformers/issues/7735
        # once this is fixed in transformers, we should remove this hacky fix
        self.tokenizer.save_vocabulary(".")
        slow_tokenizer = transformers.BertTokenizer("./vocab.txt", **self.tokenizer.init_kwargs)
        os.remove("./vocab.txt")
        self.pipeline = transformers.QuestionAnsweringPipeline(model=self.model, tokenizer=slow_tokenizer,
                                                               device=-1 if self.model.device.type == "cpu" else
                                                               self.model.device.index)

    def predict(self, inputs=None, question=None, context=None, **kwargs):
        """
        Answer the question(s) given as inputs by using the context(s).
        Takes either transformers.SquadExample (or list of them) or lists of strings, see argument documentation.

        Parameters
        ----------
        inputs: transformers.SquadExample or a list of SquadExample
            One or several SquadExample containing the question and context.
        question: str or List[str]
            One or several question(s) (must be used in conjunction with the context argument).
        context: str or List[str]
            One or several context(s) associated with the question(s) (must be used in conjunction with the question
            argument).
        kwargs include:
        topk: int, defaults to 1
            The number of answers to return (will be chosen by order of likelihood).
        doc_stride: int, defaults to 128
            If the context is too long to fit with the question for the model, it will be split in several chunks with
            some overlap. This argument controls the size of that overlap.
        max_answer_len: int, defaults to 15
            The maximum length of predicted answers (e.g., only answers with a shorter length are considered).
        max_seq_len: int, defaults to 384
            The maximum length of the total sentence (context + question) after tokenization. The context will be split
            in several chunks (using doc_stride) if needed.
        max_question_len: int, defaults to 64
            The maximum length of the question after tokenization. It will be truncated if needed.
        handle_impossible_answer: bool, defaults to False
            Whether or not we accept impossible as an answer.

        Returns
        -------
        Returns a list of answers, one for each question. If only one question has been passed as input, returns a list
        with one dictionary.
        Each answer is a dict with:
        score: float
            The probability associated to the answer.
        start: int
            The start index of the answer (in the tokenized version of the input).
        end: int
            The end index of the answer (in the tokenized version of the input).
        answer: str
            The answer to the question.
        """
        if inputs:
            prediction = self._predict(None, data=[inputs] if isinstance(inputs, transformers.SquadExample) else inputs,
                                       **kwargs)
        elif question and context:
            prediction = self._predict(None, question=question, context=context, **kwargs)
        else:
            raise RuntimeError("Must provide either SquadExample or both question and context to predict function.")
        if not isinstance(prediction, list):
            prediction = [prediction]
        return prediction

    @classmethod
    def load(cls, pretrained_tokenizer, pretrained_model, device=None):
        """
        Load a mangoes BERTForQuestionAnswering object from saved tokenizer and model files.
        This is the preferred way to initialize this class, as the base BERT model should be pretrained.
        This function follows the transformers library way of loading pretrained models for fine-tuning, which allows
        for the following use-cases:
            - Load just the base BERT model, for use in fine-tuning.
            - Load the base BERT model and the fine tuned heads, for use in inference or further fine-tuning.

        Parameters
        ----------
        pretrained_tokenizer: str
            Either:
                - A string with the `shortcut name` of a pretrained model to load from cache or download, e.g.,
                  ``bert-base-uncased``.
                - A string with the `identifier name` of a pretrained model that was user-uploaded to our S3, e.g.,
                  ``dbmdz/bert-base-german-cased``.
                - A path to a `directory` containing model weights saved using
                  :func:`~save_pretrained`, e.g., ``./my_model_directory/``.
        pretrained_model: str
            Either:
                - A string with the `shortcut name` of a pretrained model to load from cache or download, e.g.,
                  ``bert-base-uncased``.
                - A string with the `identifier name` of a pretrained model that was user-uploaded to our S3, e.g.,
                  ``dbmdz/bert-base-german-cased``.
                - A path to a `directory` containing model weights saved using
                  :func:`~save_pretrained`, e.g., ``./my_model_directory/``.
        device: int, optional, defaults to None
            if -1, use cpu, if >= 0, use CUDA device number. If None, will use GPU if available
        """
        model = transformers.BertForQuestionAnswering.from_pretrained(pretrained_model)
        model_object = cls(pretrained_tokenizer, device=device)
        model_object.model = model
        model_object.model.eval()
        return model_object

    def train(self, output_dir=None, train_question_texts=None, eval_question_texts=None, train_context_texts=None,
              eval_context_texts=None, train_answer_texts=None, eval_answer_texts=None, train_start_indices=None,
              eval_start_indices=None, max_seq_length=384, doc_stride=128, max_query_length=64, freeze_base=False,
              task_learn_rate=None, collator=None, train_dataset=None, eval_dataset=None, compute_metrics=None,
              trainer=None, **training_args):
        """
        Fine tune a BERT model on a question answering dataset

        Parameters
        ----------
        output_dir: str
            Path to the output directory where the model predictions and checkpoints will be written. Used to
            instantiate Trainer if trainer argument is None.
        train_question_texts, eval_question_texts: list of str
            The texts corresponding to the questions
        train_context_texts, eval_context_texts: list of str
            The texts corresponding to the contexts
        train_answer_texts, eval_answer_texts: list of str
            The texts corresponding to the answers
        train_start_indices, eval_start_indices: list of int
            The character positions of the start of the answers
        max_seq_length:int
            The maximum total input sequence length after tokenization.
        doc_stride: int
            When splitting up a long document into chunks, how much stride to take between chunks.
        max_query_length: int
            The maximum number of tokens for the question.
        freeze_base: Boolean
            Whether to freeze the weights of the base BERT model, so training only changes the task head weights.
            If true, the requires_grad flag for parameters of the base model will be set to false before training.
        task_learn_rate: float
            Learning rate to be used for task specific parameters, (base parameters will use the normal, ie already
            defined in **training_args, learning rate). If None, all parameters will use the same normal learning rate.
        collator: Transformers.DataCollator
            custom collator to use
        train_dataset, eval_dataset: torch.Dataset
            instantiated custom dataset object
        compute_metrics: function
            The function that will be used to compute metrics at evaluation. Must return a dictionary string to metric
            values. Used by the trainer, see https://huggingface.co/transformers/training.html#trainer for more info.
        trainer: Transformers.Trainer
            custom instantiated trainer to use
        training_args:
            keyword arguments for training. For complete list, see
            https://huggingface.co/transformers/main_classes/trainer.html#transformers.TrainingArguments
        """
        if freeze_base:
            freeze_base_layers(self.model)
        if not trainer:
            if not output_dir:
                raise RuntimeError("Must provide output directory argument to train() method if trainer argument is "
                                   "None")
            if not train_dataset:
                if not train_question_texts or not train_context_texts or not train_answer_texts or \
                        not train_start_indices:
                    raise RuntimeError("Incomplete training data provided to train method")
                train_dataset = MangoesQuestionAnsweringDataset(self.tokenizer, train_question_texts,
                                                                train_context_texts, train_answer_texts,
                                                                train_start_indices, max_seq_length, doc_stride,
                                                                max_query_length)
            if eval_question_texts and eval_context_texts and eval_answer_texts and eval_start_indices and \
                    not eval_dataset:
                eval_dataset = MangoesQuestionAnsweringDataset(self.tokenizer, eval_question_texts,
                                                               eval_context_texts, eval_answer_texts,
                                                               eval_start_indices, max_seq_length, doc_stride,
                                                               max_query_length)
            if eval_dataset is not None and "evaluation_strategy" not in training_args:
                training_args["evaluation_strategy"] = "epoch"
            train_args = transformers.TrainingArguments(output_dir=output_dir, **training_args)
            if task_learn_rate is not None:
                trainer = MultipleLearnRateFineTuneTrainer(task_learn_rate=task_learn_rate, model=self.model,
                                                           args=train_args, train_dataset=train_dataset,
                                                           data_collator=collator, eval_dataset=eval_dataset,
                                                           tokenizer=self.tokenizer, compute_metrics=compute_metrics)
            else:
                trainer = transformers.Trainer(self.model, args=train_args, train_dataset=train_dataset,
                                               data_collator=collator, eval_dataset=eval_dataset,
                                               tokenizer=self.tokenizer, compute_metrics=compute_metrics)
        trainer.train()
        self.model.eval()


class BERTForMultipleChoice(TransformerModel):
    """
    BERT model with a multiple choice classification head on top (a linear layer on top of the pooled output and a
    softmax) e.g. for RocStories/SWAG tasks.

    For information on how multiple choice datasets should be formatted for BERT fine-tuning, see this explanation:
    https://github.com/google-research/bert/issues/38

    And this link for explanation of Huggingface's multiple choice models:
    https://github.com/huggingface/transformers/issues/7701#issuecomment-707149546

    :var self.tokenizer: transformers.BertTokenizerFast object,
        see https://huggingface.co/transformers/model_doc/bert.html#berttokenizerfast for more documentation
    :var self.model: transformers.BertForMultipleChoice object,
        see https://huggingface.co/transformers/model_doc/bert.html#bertformultiplechoice

    The init function should rarely be used to directly instantiate this object, as it will initialize an untrained
    base BERT model; use the :func:`~load` function instead, as this can be used to load a pretrained base BERT
    model for fine-tuning (or a pretrained base and fine-tuning head(s) for inference or further fine-tuning).
    This follows the transformers library framework for saving/loading.

    Parameters
    ----------
    pretrained_tokenizer: str
        Either:
            - A string with the `shortcut name` of a pretrained tokenizer to load from cache or download, e.g.,
              ``bert-base-uncased``.
            - A string with the `identifier name` of a pretrained tokenizer that was user-uploaded to our S3, e.g.,
              ``dbmdz/bert-base-german-cased``.
            - A path to a `directory` containing tokenizer vocab or file saved using
              :func:`tokenizer.save_model` or :func:`tokenizer.save`, e.g., ``./my_model_directory/``.
    device: int, optional, defaults to None
        if -1, use cpu, if >= 0, use CUDA device number. If None, will use GPU if available
    **keyword_args include arguments passed to transformers.BertConfig (see mangoes.modeling.BERTBase docstring)
    """
    def __init__(self, pretrained_tokenizer, device=None, **keyword_args):
        TransformerModel.__init__(self, device)
        self.tokenizer = transformers.BertTokenizerFast.from_pretrained(pretrained_tokenizer)
        keyword_args["vocab_size"] = len(self.tokenizer.get_vocab())
        config = transformers.BertConfig(**keyword_args)
        self.model = transformers.BertForMultipleChoice(config)

    def generate_outputs(self, questions, choices, pre_tokenized=False, output_attentions=False,
                         output_hidden_states=False, word_embeddings=False):
        """
        Tokenize context and choices and pass them through the BERT model and MC head,
        optionally outputting hidden states or attention matrices.
        Works for a single question/set of choices or a batch. If a single question/set of choices is given, a batch of
        size 1 will be created.

        Follows these explanations for packing MC data and sending it through the model:
            https://github.com/google-research/bert/issues/38
            https://github.com/huggingface/transformers/issues/7701#issuecomment-7071495

        Parameters
        ----------
        questions: str or List[str] if pre_tokenized=False, else List[str] or List[List[str]]
            The question text. This can include the context together with a question, or (in the case of some datasets
            such as the SWAG dataset) just the context if there is no direct question. Can be a single question or list
            of questions.
        choices: List[str] or List[List[str]] if pre_tokenized=False, else List[List[str]] or List[List[List[str]]]
            The choices text. One instance of choices should be a list of strings (if not pre-tokenized) or a list of
            list of strings (if pre-tokenized). Can be a single choice instance or multiple.
            If batch is passed in (ie more than one question), assumes all questions have same number of choices.
        pre_tokenized: Boolean
            Whether or not the input text is pre-tokenized (ie, split on spaces)
        output_attentions: Boolean, optional, defaults to False
            Whether or not to return the attentions tensors of all attention layers.
        output_hidden_states: Boolean, optional, defaults to False
            Whether or not to return the hidden states of all layers.
        word_embeddings: Boolean
            Whether or not to filter special token embeddings and average subword embeddings (hidden states) into word
            embeddings. This functionality is not available for this task class. Use the feature extraction class
            instead.

        Returns
        -------
        Dict containing (note that if single question/set of choices is passed as input, batch size will be 1):
        logits: Tensor of shape (batch_size, num_choices)
            Classification scores (before SoftMax). If batch
        offset_mappings: Tensor of shape (batch_size, num_choices, sequence_length, 2)
            Tensor containing (char_start, char_end) for each token, giving index into input strings of start and end
            character for each token. If input is pre-tokenized, start and end index maps into associated word. Note
            that special tokens are included with 0 for start and end indices, as these don't map into input text
            because they are added inside the function. Offset mappings for both questions and choices are merged to one
            row, but indices still align to them separately.
        hidden_states: Tuple (one for each layer) of torch.FloatTensor of size
            (batch_size, num_choices, sequence_length, hidden_size)
            Hidden-states of the model at the output of each layer plus the initial embedding outputs.
            Only returned if output_hidden_states is True
        attentions: Tuple of torch.FloatTensor (one for each layer) of shape (batch_size, num_choices, num_heads,
            sequence_length, sequence_length).
            Attentions weights after the attention softmax, used to compute the weighted average in the self-attention
            heads.
            Only returned if output_attentions is True
        """
        # make batch of 1 if single question
        if isinstance(questions, str) or (isinstance(questions[0], str) and pre_tokenized):
            questions = [questions]
        if isinstance(choices[0], str) or (isinstance(choices[0][0], str) and pre_tokenized):
            choices = [choices]
        if len(questions) != len(choices):
            raise RuntimeError("Number of questions does not match number of sets of choices, "
                               "please checkxt classification inputs. Refer to the method documentation for input "
                               "format")
        batch_size = len(questions)
        num_choices = len(choices[0])
        for i in range(batch_size):
            questions[i] = [questions[i]] * num_choices
        # flatten into one batch of (batch_size * num_choices) sequences
        questions = [single_choice_question for question in questions for single_choice_question in question]
        choices = [single_choice for choice_list in choices for single_choice in choice_list]

        inputs = self.tokenizer(
            text=questions,
            text_pair=choices,
            padding=True,
            is_split_into_words=pre_tokenized,
            return_tensors="pt",
            return_offsets_mapping=True,
        )
        sequence_length = inputs["input_ids"].size(-1)
        inputs["input_ids"] = inputs["input_ids"].view(batch_size, num_choices, -1)
        inputs["token_type_ids"] = inputs["token_type_ids"].view(batch_size, num_choices, -1)
        inputs["attention_mask"] = inputs["attention_mask"].view(batch_size, num_choices, -1)
        offset_mappings = inputs.pop('offset_mapping')
        offset_mappings = offset_mappings.view(batch_size, num_choices, -1, 2)

        self.model.eval()
        if word_embeddings:
            warnings.warn("Word embedding consolidation not available for Multiple Choice model. Consider using the "
                          "mangoes.modeling.BERTBase class for feature extractions")
        with torch.no_grad():
            if self.model_device.type == "cuda":
                inputs = {name: tensor.to(self.model_device) for name, tensor in inputs.items()}
                if not next(self.model.parameters()).is_cuda:
                    self.model.to(self.model_device)
            results = self.model.forward(**inputs, return_dict=True, output_attentions=output_attentions,
                                         output_hidden_states=output_hidden_states)
        if output_hidden_states:
            resized_hidden_states = []
            for i in range(len(results["hidden_states"])):
                resized_hidden_states.append(results["hidden_states"][i]
                                             .view(batch_size, num_choices, sequence_length, -1))
            results["hidden_states"] = tuple(resized_hidden_states)
        if output_attentions:
            resized_attentions = []
            for i in range(len(results["attentions"])):
                resized_attentions.append(results["attentions"][i]
                                          .view(batch_size, num_choices, -1, sequence_length, sequence_length))
            results["attentions"] = tuple(resized_attentions)
        results["offset_mappings"] = offset_mappings
        return results

    def predict(self, questions, choices, pre_tokenized=False):
        """
        Predicts the answer to the question(s) out of the possible choices.

        Parameters
        ----------
        questions: str or List[str] if pre_tokenized=False, else List[str] or List[List[str]]
            The question text. This can include the context together with a question, or (in the case of some datasets
            such as the SWAG dataset) just the context if there is no direct question. Can be a single question or list
            of questions.
        choices: List[str] or List[List[str]] if pre_tokenized=False, else List[List[str]] or List[List[List[str]]]
            The choices text. One instance of choices should be a list of strings (if not pre-tokenized) or a list of
            list of strings (if pre-tokenized). Can be a single choice instance or multiple.
            If batch is passed in (ie more than one question), assumes all questions have same number of choices.
        pre_tokenized: Boolean
            Whether or not the input text is pre-tokenized (ie, split on spaces)

        Returns
        -------
        List of answer prediction dicts, one for each question (returns list of length 1 if single question is passed as
        input. Answer prediction dicts include:
        score: float
            The probability associated to the answer.
        answer_index: int
            The index of the predicted choice.
        answer_text: str (if not pre_tokenized) or List[str] (if pre_tokenized)
            The text corresponding to the predicted answer.
        """
        outputs = self.generate_outputs(questions, choices, pre_tokenized)
        if isinstance(choices[0], str) or (pre_tokenized and isinstance(choices[0][0], str)):
            choices = [choices]
        outputs["logits"] = F.softmax(outputs["logits"], dim=-1)
        predictions = []
        for i in range(len(outputs["logits"])):
            answer_dict = {"answer_index": torch.argmax(outputs["logits"][i]).item()}
            answer_dict["score"] = outputs["logits"][i][answer_dict["answer_index"]].item()
            answer_dict["answer_text"] = choices[i][answer_dict["answer_index"]]
            predictions.append(answer_dict)
        return predictions

    @classmethod
    def load(cls, pretrained_tokenizer, pretrained_model, device=None):
        """
        Load a mangoes BERTForMultipleChoice object from saved tokenizer and model files.
        This is the preferred way to initialize this class, as the base BERT model should be pretrained.
        This function follows the transformers library way of loading pretrained models for fine-tuning, which allows
        for the following use-cases:
            - Load just the base BERT model, for use in fine-tuning.
            - Load the base BERT model and the fine tuned heads, for use in inference or further fine-tuning.

        Parameters
        ----------
        pretrained_tokenizer: str
            Either:
                - A string with the `shortcut name` of a pretrained model to load from cache or download, e.g.,
                  ``bert-base-uncased``.
                - A string with the `identifier name` of a pretrained model that was user-uploaded to our S3, e.g.,
                  ``dbmdz/bert-base-german-cased``.
                - A path to a `directory` containing model weights saved using
                  :func:`~save_pretrained`, e.g., ``./my_model_directory/``.
        pretrained_model: str
            Either:
                - A string with the `shortcut name` of a pretrained model to load from cache or download, e.g.,
                  ``bert-base-uncased``.
                - A string with the `identifier name` of a pretrained model that was user-uploaded to our S3, e.g.,
                  ``dbmdz/bert-base-german-cased``.
                - A path to a `directory` containing model weights saved using
                  :func:`~save_pretrained`, e.g., ``./my_model_directory/``.
        device: int, optional, defaults to None
            if -1, use cpu, if >= 0, use CUDA device number. If None, will use GPU if available
        """
        model = transformers.BertForMultipleChoice.from_pretrained(pretrained_model)
        model_object = cls(pretrained_tokenizer, device=device)
        model_object.model = model
        model_object.model.eval()
        return model_object

    def train(self, output_dir=None, train_question_texts=None, eval_question_texts=None, train_choices_texts=None,
              eval_choices_texts=None, train_labels=None, eval_labels=None, max_len=384, freeze_base=False,
              task_learn_rate=None, collator=None, train_dataset=None, eval_dataset=None, compute_metrics=None,
              trainer=None, **training_args):
        """
        Fine tune a BERT model on a multiple choice dataset

        Parameters
        ----------
        output_dir: str
            Path to the output directory where the model predictions and checkpoints will be written. Used to
            instantiate Trainer if trainer argument is None.
        train_question_texts, eval_question_texts: list of str
            The texts corresponding to the questions/contexts.
        train_choices_texts, eval_choices_texts: list of str
            The texts corresponding to the answer choices
        train_labels, eval_labels: list of int
            The indices of the correct answers
        max_len:int
            The maximum total input sequence length after tokenization. Note that if a question answer pair sequence is
            longer than this length, it will be truncated.
        freeze_base: Boolean
            Whether to freeze the weights of the base BERT model, so training only changes the task head weights.
            If true, the requires_grad flag for parameters of the base model will be set to false before training.
        task_learn_rate: float
            Learning rate to be used for task specific parameters, (base parameters will use the normal, ie already
            defined in **training_args, learning rate). If None, all parameters will use the same normal learning rate.
        collator: Transformers.DataCollator
            custom collator to use
        train_dataset, eval_dataset: torch.Dataset
            instantiated custom dataset object
        compute_metrics: function
            The function that will be used to compute metrics at evaluation. Must return a dictionary string to metric
            values. Used by the trainer, see https://huggingface.co/transformers/training.html#trainer for more info.
        trainer: Transformers.Trainer
            custom instantiated trainer to use
        training_args:
            keyword arguments for training. For complete list, see
            https://huggingface.co/transformers/main_classes/trainer.html#transformers.TrainingArguments
        """
        if freeze_base:
            freeze_base_layers(self.model)
        if not trainer:
            if not output_dir:
                raise RuntimeError("Must provide output directory argument to train() method if trainer argument is "
                                   "None")
            if not train_dataset:
                if not train_question_texts or not train_choices_texts or not train_labels:
                    raise RuntimeError("Incomplete training data provided to train method")
                train_dataset = MangoesMultipleChoiceDataset(self.tokenizer, train_question_texts, train_choices_texts,
                                                             train_labels, max_len)
            if eval_question_texts and eval_choices_texts and eval_labels and not eval_dataset:
                eval_dataset = MangoesMultipleChoiceDataset(self.tokenizer, eval_question_texts, eval_choices_texts,
                                                            eval_labels, max_len)
            if eval_dataset is not None and "evaluation_strategy" not in training_args:
                training_args["evaluation_strategy"] = "epoch"
            train_args = transformers.TrainingArguments(output_dir=output_dir, **training_args)
            if task_learn_rate is not None:
                trainer = MultipleLearnRateFineTuneTrainer(task_learn_rate, model=self.model, args=train_args,
                                                           data_collator=collator, train_dataset=train_dataset,
                                                           eval_dataset=eval_dataset, tokenizer=self.tokenizer,
                                                           compute_metrics=compute_metrics)
            else:
                trainer = transformers.Trainer(self.model, args=train_args, train_dataset=train_dataset,
                                               data_collator=collator, eval_dataset=eval_dataset,
                                               tokenizer=self.tokenizer, compute_metrics=compute_metrics)
        trainer.train()
        self.model.eval()


class BERTForCoreferenceResolution(TransformerModel):
    """Class for fine tuning a Bert model for the coreference resolution task.

    The base model is an implementation of the independent variant of https://arxiv.org/pdf/1908.09091.pdf, which uses
    the fine tuning procedure described in https://arxiv.org/pdf/1804.05392.pdf

    :var self.tokenizer: transformers.BertTokenizerFast object,
        see https://huggingface.co/transformers/model_doc/bert.html#berttokenizerfast for more documentation
    :var self.model: mangoes.modeling.coref.BertForCoreferenceResolutionBase object.

    The init function should rarely be used to directly instantiate this object, as it will initialize an untrained
    base BERT model; use the :func:`~load` function instead, as this can be used to load a pretrained base BERT
    model for fine-tuning (or a pretrained base and fine-tuning head(s) for inference or further fine-tuning).
    This follows the transformers library framework for saving/loading.

    Parameters
    ----------
    pretrained_tokenizer: str
        Either:
            - A string with the `shortcut name` of a pretrained tokenizer to load from cache or download, e.g.,
              ``bert-base-uncased``.
            - A string with the `identifier name` of a pretrained tokenizer that was user-uploaded to our S3, e.g.,
              ``dbmdz/bert-base-german-cased``.
            - A path to a `directory` containing tokenizer vocab or file saved using
              :func:`tokenizer.save_model` or :func:`tokenizer.save`, e.g., ``./my_model_directory/``.
    device: int, optional, defaults to None
        If -1, use cpu, if >= 0, use CUDA device number. If None, will use GPU if available.
    max_span_width: int, defaults to 30
        Maximum width (consecutive tokens) of candidate span.
    ffnn_hidden_size: int, defaults to 1000
        Size of hidden layers in dense mention scorer and slow antecedent scorer heads.
    top_span_ratio: float, defaults to 0.4
        Ratio of spans to consider after first sort on mention score.
    max_top_antecendents" int, defaults to 50
        Max number of antecedents to consider for each span after fast antecedent scorer.
    use_metadata: Boolean, defaults to False
        Whether to use metadata (speaker and genre information) in forward pass.
    metadata_feature_size: int, defaults to 20
        Size of metadata features (if using metadata)
    genres: List of string, defaults to ("bc", "bn", "mz", "nw", "pt", "tc", "wb")
        List of possible genres (if using metadata). Defaults to genres in Ontonotes dataset.
    max_training_segments: int, defaults to 5
        Maximum number of segments in one document (aka one batch).
    coref_depth: int, defaults to 2
        Depth of higher order (aka slow) antecedent scoring.
    coref_dropout: float, defaults to 0.3
        Dropout probability for head layers.
    **keyword_args include arguments passed to transformers.BertConfig (see mangoes.modeling.BERTBase docstring)
    """
    def __init__(self, pretrained_tokenizer, device=None, max_span_width=30, ffnn_hidden_size=1000,
                 top_span_ratio=0.4, max_top_antecendents=50, use_metadata=False, metadata_feature_size=20,
                 genres=("bc", "bn", "mz", "nw", "pt", "tc", "wb"), max_training_segments=5, coref_depth=2,
                 coref_dropout=0.3, **keyword_args):
        TransformerModel.__init__(self, device)
        self.tokenizer = transformers.BertTokenizerFast.from_pretrained(pretrained_tokenizer)
        keyword_args["vocab_size"] = len(self.tokenizer.get_vocab())
        config = transformers.BertConfig(**keyword_args)
        self.model = BertForCoreferenceResolutionBase(config, max_span_width=max_span_width,
                                                      ffnn_hidden_size=ffnn_hidden_size, top_span_ratio=top_span_ratio,
                                                      max_top_antecendents=max_top_antecendents,
                                                      use_metadata=use_metadata,
                                                      metadata_feature_size=metadata_feature_size,
                                                      genres=genres, max_training_segments=max_training_segments,
                                                      coref_depth=coref_depth, coref_dropout=coref_dropout)

    def predict(self, text, pre_tokenized=False, speaker_ids=None, genre=None, max_segment_len=256, max_segments=5):
        """
        Predict the co-reference clusters in text. Takes one document at a time. Internally calls generate_outputs and
        then processes the outputs.

        Parameters
        ----------
        text: str or List[str] if pre_tokenized=False, else List[str] or List[List[str]]
            The text to predict co-references for. If pre_tokenized, the text can be one sentence (list of words) or
            list of sentences (list of list of words). If not pre_tokenized, text can be one sentence (str) or list of
            sentences (list of strings).
        pre_tokenized: Boolean
            Whether or not the input text is pretokenized (ie, split on spaces). This method will still pass it through
            the tokenizer, in order to get subtokens, special characters, and attention masks.
        speaker_ids: int or List[int] if pre_tokenized=False, else List[int] or List[List[int]]
            Speaker ids for input text. If pre_tokenized, speaker_ids should be for each word in each sentence of input
            text (ie, list of int if one sentence, or list of list of int if multiple). If not pre_tokenized, speaker
            ids should be on a sentence basis, ie one int if one input sentence, or list of ints if multiple.
            Optional, needed only if the model has been trained/instantiated to accept metadata.
        genre: Int or String
            Genre of text. If string, will attempt to use the genre id mapping constructed by the model parameter to
            this object. Optional, needed only if the model has been trained/instantiated to accept metadata.
        max_segment_len: int, defaults to 256
            maximum number of sub-tokens for one segment
        max_segments: int, defaults to 5
            Maximum number of segments to return per document

        Returns
        -------
        List of dicts.
        For each found co-reference cluster, a dict with the following keys:
            cluster_tokens: List[List[str]]]
                The text spans associated with the cluster. Spans are represented by the list of tokens.
            cluster_ids: List[List[int]]
                The id spans associated with the cluster. Spans are represented by the list of token ids.
        """
        outputs = self.generate_outputs(text, pre_tokenized, speaker_ids, genre, max_segment_len, max_segments)
        top_indices = torch.argmax(outputs["top_antecedent_scores"], dim=-1, keepdim=False)

        mention_indices = []
        antecedent_indices = []
        for i in range(len(outputs["top_span_ends"])):
            if top_indices[i] > 0:
                mention_indices.append(i)
                antecedent_indices.append(outputs["top_antecedents"][i][top_indices[i] - 1].item())

        cluster_sets = []
        for i in range(len(antecedent_indices)):
            new_cluster = True
            for j in range(len(cluster_sets)):
                if mention_indices[i] in cluster_sets[j] or antecedent_indices[i] in cluster_sets[j]:
                    cluster_sets[j].add(mention_indices[i])
                    cluster_sets[j].add(antecedent_indices[i])
                    new_cluster = False
                    break
            if new_cluster:
                cluster_sets.append({mention_indices[i], antecedent_indices[i]})

        cluster_dicts = []
        for i in range(len(cluster_sets)):
            cluster_mentions = sorted(list(cluster_sets[i]))
            current_ids = []
            current_tokens = []
            for mention_index in cluster_mentions:
                current_ids.append(outputs["flattened_ids"]
                                   [outputs["top_span_starts"][mention_index]:
                                    outputs["top_span_ends"][mention_index] + 1])
                current_tokens.append(outputs["flattened_text"]
                                      [outputs["top_span_starts"][mention_index]:
                                       outputs["top_span_ends"][mention_index] + 1])
            cluster_dicts.append({"cluster_ids": current_ids, "cluster_tokens": current_tokens})
        return cluster_dicts

    def generate_outputs(self, text, pre_tokenized=False, speaker_ids=None, genre=None, output_attentions=False,
                         output_hidden_states=False, word_embeddings=False, max_segment_len=256, max_segments=5):
        """
        Pass one batch (document) worth of text through the co-reference model, outputting mention scores and token
        indices for possible spans, the indices and scores of the top antecedents for the top mention scored spans, and
        optionally the hidden states or attention matrices from the base BERT model.

        Note that this functions does not return "offset_mappings", and instead returns the flattened ids and text.

        Parameters
        ----------
        text: str or List[str] if pre_tokenized=False, else List[str] or List[List[str]]
            The text to predict co-references for. If pre_tokenized, the text can be one sentence (list of words) or
            list of sentences (list of list of words). If not pre_tokenized, text can be one sentence (str) or list of
            sentences (list of strings).
        pre_tokenized: Boolean
            Whether or not the input text is pretokenized (ie, split on spaces). This method will still pass it through
            the tokenizer, in order to get subtokens, special characters, and attention masks.
        speaker_ids: int or List[int] if pre_tokenized=False, else List[int] or List[List[int]]
            Speaker ids for input text. If pre_tokenized, speaker_ids should be for each word in each sentence of input
            text (ie, list of int if one sentence, or list of list of int if multiple). If not pre_tokenized, speaker
            ids should be on a sentence basis, ie one int if one input sentence, or list of ints if multiple.
            Optional, needed only if the model has been trained/instantiated to accept metadata.
        genre: Int or String
            Genre of text. If string, will attempt to use the genre id mapping constructed by the model parameter to
            this object. Optional, needed only if the model has been trained/instantiated to accept metadata.
        output_attentions: Boolean, optional, defaults to False
            Whether or not to return the attentions tensors of all attention layers.
        output_hidden_states: Boolean, optional, defaults to False
            Whether or not to return the hidden states of all layers.
        word_embeddings: Boolean
            Whether or not to filter special token embeddings and average subword embeddings (hidden states) into word
            embeddings. Note: this functionality is not available for this class because of consolidation of input in
            the forward pass of the model. Consider using
            mangoes.modeling.PretrainedTransformerModelForFeatureExtraction class for word-level feature extractions.
        max_segment_len: int, defaults to 256
            maximum number of sub-tokens for one segment
        max_segments: int, defaults to 5
            Maximum number of segments to return per document

        Returns
        -------
        Dict containing:
        candidate_starts: tensor of size (num_spans)
            start token indices in flattened document of candidate spans
        candidate_ends: tensor of size (num_spans)
            end token indices in flattened document of candidate spans
        candidate_mention_scores: tensor of size (num_spans)
            mention scores for each candidate span
        top_span_starts: tensor of size (num_top_spans)
            start token indices in flattened document of candidate spans with top mention scores
        top_span_ends: tensor of size (num_top_spans)
            end token indices in flattened document of candidate spans with top mention scores
        top_antecedents: tensor of shape (num_top_spans, antecedent_candidates)
            indices in top span candidates of top antecedents for each mention
        top_antecedent_scores: tensor of shape (num_top_spans, 1 + antecedent_candidates)
            final antecedent scores of top antecedents for each mention. The dummy score (for not a co-reference) is
            inserted at the start of each row. Thus, the score for top_antecedents[0][0] is top_antecedent_scores[0][1].
            The span for first candidate is top_span_starts[0] to top_span_ends[0]. The span for the first top
            antecedent for the first candidate is top_span_starts[top_antecedents[0][0]] to
            top_span_ends[top_antecedents[0][0]].
        flattened_ids: tensor of shape (num_tokens)
            flattened ids of input sentences. The start and end candidate and span indices map into this tensor.
        flattened_text: tensor of shape (num_tokens)
            flattened tokens of input sentences. The start and end candidate and span indices map into this tensor.
        hidden_states: Tuple (one for each layer) of torch.FloatTensor (batch_size, sequence_length, hidden_size).
            Hidden-states of the model at the output of each layer plus the initial embedding outputs.
            Only returned if output_hidden_states is True
        attentions: Tuple of torch.FloatTensor (one for each layer) of shape (batch_size, num_heads, sequence_length,
            sequence_length).
            Attentions weights after the attention softmax, used to compute the weighted average in the self-attention
            heads.
            Only returned if output_attentions is True
        """
        if self.model.use_metadata:
            if not (speaker_ids and genre):
                raise RuntimeError("Co-reference model has been trained/initiated to use metadata, please provide "
                                   "speaker_ids argument and genre arguments.")
            if isinstance(genre, str):
                try:
                    genre = self.model.genres[genre]
                except KeyError:
                    raise RuntimeError("The genre argument passed to generate_outputs is not in the model's genre id "
                                       "mapping dictionary.")
            elif not isinstance(genre, int):
                raise RuntimeError("The genre argument must be a string or int.")
            if isinstance(text, list):
                if not len(speaker_ids) == len(text):
                    raise RuntimeError("Lengths of speaker ids and text arguments are different. See docstring of "
                                       "generate_outputs for appropriate input format.")
        # if single sentence, wrap in list
        if isinstance(text, str) or (pre_tokenized and isinstance(text[0], str)):
            text = [text]
            if self.model.use_metadata:
                speaker_ids = [speaker_ids]

        subtoken_ids = []
        subtoken_speakers = []
        subtoken_offset_mappings = []
        for i in range(len(text)):
            encoding = self.tokenizer(text[i], add_special_tokens=False, is_split_into_words=pre_tokenized,
                                      return_offsets_mapping=True)
            subtoken_ids.append(encoding["input_ids"])
            subtoken_offset_mappings.append(encoding["offset_mapping"])
            if self.model.use_metadata:
                if pre_tokenized:
                    subtoken_speakers.append(MangoesCoreferenceDataset.get_subtoken_data(speaker_ids[i],
                                                                                         encoding["offset_mapping"]))
                else:
                    subtoken_speakers.append([speaker_ids[i] for _ in range(len(encoding["input_ids"]))])

        current_segment_ids = []
        current_segment_speaker_ids = []
        current_sentence_map = []
        segments_ids = []
        segments_speakers = []
        segments_attention_mask = []
        sentence_map = []
        for j in range(len(subtoken_ids)):
            if len(current_segment_ids) + len(subtoken_ids[j]) <= max_segment_len - 2:
                current_segment_ids += subtoken_ids[j]
                current_sentence_map += [j] * len(subtoken_ids[j])
                if self.model.use_metadata:
                    current_segment_speaker_ids += subtoken_speakers[j]
            else:
                # segments contain cls and sep special tokens at beginning and end for BERT processing
                if len(current_segment_ids) > 0:
                    segments_ids.append(MangoesCoreferenceDataset.pad_list([self.tokenizer.cls_token_id] +
                                                                           current_segment_ids +
                                                                           [self.tokenizer.sep_token_id],
                                                                           max_segment_len))
                    segments_attention_mask.append(MangoesCoreferenceDataset.pad_list([1] *
                                                                                      (len(current_segment_ids) + 2),
                                                                                      max_segment_len))
                    sentence_map += [current_sentence_map[0]] + current_sentence_map + [current_sentence_map[-1]]
                    if self.model.use_metadata:
                        segments_speakers.append(MangoesCoreferenceDataset.pad_list([-1] + current_segment_speaker_ids +
                                                                                    [-1], max_segment_len))
                if len(subtoken_ids[j]) > max_segment_len - 2:
                    # if sentence j is longer than max_seq_len, create segment out of as much as possible,
                    # then remove these from sentence j and continue
                    segment_stop_index = max_segment_len - 2
                    while subtoken_offset_mappings[j][segment_stop_index-1][0] > 0 or \
                            subtoken_offset_mappings[j][segment_stop_index][0] > 0:
                        # if breaking sentence in the middle of a token, truncate so whole token is in next segment
                        segment_stop_index -= 1
                    segments_ids.append(MangoesCoreferenceDataset.pad_list([self.tokenizer.cls_token_id] +
                                                                           subtoken_ids[j][:segment_stop_index] +
                                                                           [self.tokenizer.sep_token_id],
                                                                           max_segment_len))
                    segments_attention_mask.append(
                        MangoesCoreferenceDataset.pad_list([1] * (segment_stop_index + 2), max_segment_len))
                    sentence_map += [j] * (segment_stop_index + 2)
                    if self.model.use_metadata:
                        segments_speakers.append(MangoesCoreferenceDataset.pad_list([-1] +
                                                                                    subtoken_speakers[j]
                                                                                    [:segment_stop_index] +
                                                                                    [-1], max_segment_len))
                    # remove already added data
                    subtoken_ids[j] = subtoken_ids[j][segment_stop_index:]
                    if self.model.use_metadata:
                        subtoken_speakers[j] = subtoken_speakers[j][segment_stop_index:]
                current_segment_ids = subtoken_ids[j]
                current_sentence_map = [j] * len(subtoken_ids[j])
                if self.model.use_metadata:
                    current_segment_speaker_ids = subtoken_speakers[j]
        # get last segment
        segments_ids.append(MangoesCoreferenceDataset.pad_list([self.tokenizer.cls_token_id] + current_segment_ids +
                                                               [self.tokenizer.sep_token_id], max_segment_len))
        segments_attention_mask.append(MangoesCoreferenceDataset.pad_list([1] * (len(current_segment_ids) + 2),
                                                                          max_segment_len))
        sentence_map += [current_sentence_map[0]] + current_sentence_map + [current_sentence_map[-1]]
        if self.model.use_metadata:
            segments_speakers.append(MangoesCoreferenceDataset.pad_list([-1] + current_segment_speaker_ids + [-1],
                                                                        max_segment_len))
        ids = torch.as_tensor(segments_ids)
        attention_mask = torch.as_tensor(segments_attention_mask)
        sentence_map = torch.as_tensor(sentence_map)
        if self.model.use_metadata:
            speaker_ids = torch.as_tensor(segments_speakers)
            genre = torch.as_tensor(genre)
        if len(ids) > max_segments:
            warnings.warn('Input text to generate_outputs exceeds maximum segments; truncating.', RuntimeWarning)
            ids = ids[:max_segments]
            attention_mask = attention_mask[:max_segments]
            num_tokens = attention_mask.sum()
            sentence_map = sentence_map[:num_tokens]
            if self.model.use_metadata:
                speaker_ids = speaker_ids[:max_segments]
        inputs = {
            "input_ids": ids,
            "attention_mask": attention_mask,
            "sentence_map": sentence_map,
        }
        if self.model.use_metadata:
            inputs.update({"speaker_ids": speaker_ids, "genre": genre})
        self.model.eval()
        with torch.no_grad():
            if self.model_device.type == "cuda":
                inputs = {name: tensor.to(self.model_device) for name, tensor in inputs.items()}
                if not next(self.model.parameters()).is_cuda:
                    self.model.to(self.model_device)
            results = self.model.forward(**inputs, return_dict=True,
                                         output_attentions=output_attentions,
                                         output_hidden_states=output_hidden_states)
        if word_embeddings:
            warnings.warn("Word embedding consolidation not available for co-reference model. Consider using the"
                          "mangoes.modeling.BERTBase class for feature extractions.")
        results = dict(results)
        results["flattened_text"] = self.tokenizer.convert_ids_to_tokens(results["flattened_ids"])
        return results

    @classmethod
    def load(cls, pretrained_tokenizer, pretrained_model, device=None, max_span_width=None, ffnn_hidden_size=None,
             top_span_ratio=None, max_top_antecendents=None, use_metadata=None, metadata_feature_size=None,
             genres=None, max_training_segments=None, coref_depth=None, coref_dropout=None):
        """
        Load a mangoes BERTForCoreferenceResolution object from saved tokenizer and model files.
        This is the preferred way to initialize this class, as the base BERT model should be pretrained.
        This function follows the transformers library framework of loading pretrained models for fine-tuning, which
        allows for the following use-cases:
            1. Load just the base BERT model, for use in fine-tuning.
            2. Load the base BERT model and the fine tuned heads, for use in inference or further fine-tuning.
        If only loading the base BERT model, one should provide the initialization arguments for the fine-tuning
        architecture. These arguments will only be used if only loading the base BERT model, and will be ignored
        otherwise.

        Parameters
        ----------
        pretrained_tokenizer: str
            Either:
                - A string with the `shortcut name` of a pretrained model to load from cache or download, e.g.,
                  ``bert-base-uncased``.
                - A string with the `identifier name` of a pretrained model that was user-uploaded to our S3, e.g.,
                  ``dbmdz/bert-base-german-cased``.
                - A path to a `directory` containing model weights saved using
                  :func:`~save_pretrained`, e.g., ``./my_model_directory/``.
        pretrained_model: str
            - A path to a `directory` containing model weights saved using
              :func:`~save_pretrained`, e.g., ``./my_model_directory/``.
        device: int, optional, defaults to None
            if -1, use cpu, if >= 0, use CUDA device number. If None, will use GPU if available
        Fine-tuning architecture arguments (only used if only loading base BERT model, defaults to None, which will use
        initialization default values (see init function)):
            max_span_width: int, defaults to 30
                Maximum width (consecutive tokens) of candidate span.
            ffnn_hidden_size: int
                Size of hidden layers in dense mention scorer and slow antecedent scorer heads.
            top_span_ratio: float
                Ratio of spans to consider after first sort on mention score.
            max_top_antecendents: int
                Max number of antecedents to consider for each span after fast antecedent scorer.
            use_metadata: Boolean
                Whether to use metadata (speaker and genre information) in forward pass.
            metadata_feature_size: int
                Size of metadata features (if using metadata)
            genres: List of string
                List of possible genres (if using metadata). Defaults to genres in Ontonotes dataset.
            max_training_segments: int
                Maximum number of segments in one document (aka one batch).
            coref_depth: int
                Depth of higher order (aka slow) antecedent scoring.
            coref_dropout: float, defaults to 0.3
                Dropout probability for head layers.
        """
        model, loading_info = BertForCoreferenceResolutionBase.from_pretrained(pretrained_model,
                                                                               output_loading_info=True)
        if len(loading_info["missing_keys"]) > 0:
            # if only loading base BERT model
            fine_tuning_arch_keyword_args = {}
            if max_span_width:
                fine_tuning_arch_keyword_args["max_span_width"] = max_span_width
            if ffnn_hidden_size:
                fine_tuning_arch_keyword_args["ffnn_hidden_size"] = ffnn_hidden_size
            if top_span_ratio:
                fine_tuning_arch_keyword_args["top_span_ratio"] = top_span_ratio
            if max_top_antecendents:
                fine_tuning_arch_keyword_args["max_top_antecendents"] = max_top_antecendents
            if use_metadata:
                fine_tuning_arch_keyword_args["use_metadata"] = use_metadata
            if metadata_feature_size:
                fine_tuning_arch_keyword_args["metadata_feature_size"] = metadata_feature_size
            if genres:
                fine_tuning_arch_keyword_args["genres"] = genres
            if max_training_segments:
                fine_tuning_arch_keyword_args["max_training_segments"] = max_training_segments
            if coref_depth:
                fine_tuning_arch_keyword_args["coref_depth"] = coref_depth
            if coref_dropout:
                fine_tuning_arch_keyword_args["coref_dropout"] = coref_dropout
            # bit of a hacky way to get around loading weirdness but still follow transformers framework:
            base_config = model.config.to_dict()
            _ = base_config.pop("task_specific_params", None)
            model_object = cls(pretrained_tokenizer, device=device, **fine_tuning_arch_keyword_args,
                               **base_config)
            task_specific_params = model_object.model.config.task_specific_params
            model_object.model.bert = model.bert
            model_object.model.config = model.config
            model_object.model.config.task_specific_params = task_specific_params
        else:
            model_object = cls(pretrained_tokenizer, device=device)
            model_object.model = model
        model_object.model.eval()
        return model_object

    def train(self, output_dir=None, max_segment_len=256, max_segments=5, freeze_base=False, task_learn_rate=None,
              train_documents=None, train_cluster_ids=None, train_speaker_ids=None, train_genres=None,
              eval_documents=None, eval_cluster_ids=None, eval_speaker_ids=None, eval_genres=None,
              train_dataset=None, eval_dataset=None, compute_metrics=None, trainer=None, **training_args):
        """
        Fine tune a BERT model on a co-reference resolution dataset.
        Users can input the raw coreference data and a torch dataset will be created, or they can input already
        instantiated dataset(s), or an already instantiated trainer.

        Note that this implementation is based on the "independent" variant of the method introduced in
        https://arxiv.org/pdf/1908.09091.pdf, thus the batch size will be 1 (1 document per batch, with multiple
        segments per document), and a specific collator function will be used.

        Parameters
        ----------
        output_dir: str
            Path to the output directory where the model predictions and checkpoints will be written. Used to
            instantiate Trainer if trainer argument is None.
            Optional: needed if trainer is not provided
        max_segment_len: int, defaults to 256
            maximum number of sub-tokens for one segment
        max_segments: int, defaults to 5
            Maximum number of segments to return per document
        freeze_base: Boolean
            Whether to freeze the weights of the base BERT model, so training only changes the task head weights.
            If true, the requires_grad flag for parameters of the base model will be set to false before training.
        task_learn_rate: float
            Learning rate to be used for task specific parameters, (base parameters will use the normal, ie already
            defined in **training_args, learning rate). If None, all parameters will use the same normal learning rate.
        train_documents: List of Lists of Lists of strings
            Optional: needed if train_dataset or trainer is not provided
            Text for each document. As cluster ids are labeled by word, a document is a list of sentences. One
            sentence is a list of words (ie already split on whitespace/punctuation)
        train_cluster_ids: List of Lists of Lists of (ints or Tuple(int, int))
            Optional: needed if train_dataset or trainer is not provided
            Cluster ids for each word in documents argument. Assumes words that aren't mentions have either None or -1
            as id. In the case where a word belongs to two different spans (with different cluster ids), the cluster id
            for word should be a tuple of ints corresponding to the different cluster ids.
        train_speaker_ids: List of Lists of Lists of ints
            Optional: needed if train_dataset or trainer is not provided and model is using metadata
            Speaker id for each word in documents. Assumes positive ids (special tokens (such as [CLS] and [SEP] that
            are added at beginning and end of segments) will be assigned speaker ids of -1)
        train_genres: List of ints
            Optional: needed if train_dataset or trainer is not provided and model is using metadata
            Genre id for each document
        eval_documents: List of Lists of Lists of strings
            Optional: needed if train_dataset or trainer is not provided
            Text for each document. As cluster ids are labeled by word, a document is a list of sentences. One
            sentence is a list of words (ie already split on whitespace/punctuation)
        eval_cluster_ids: List of Lists of Lists of (ints or Tuple(int, int))
            Optional: needed if train_dataset or trainer is not provided
            Cluster ids for each word in documents argument. Assumes words that aren't mentions have either None or -1
            as id. In the case where a word belongs to two different spans (with different cluster ids), the cluster id
            for word should be a tuple of ints corresponding to the different cluster ids.
        eval_speaker_ids: List of Lists of Lists of ints
            Optional: needed if train_dataset or trainer is not provided and model is using metadata
            Speaker id for each word in documents. Assumes positive ids (special tokens (such as [CLS] and [SEP] that
            are added at beginning and end of segments) will be assigned speaker ids of -1)
        eval_genres: List of ints
            Optional: needed if train_dataset or trainer is not provided and model is using metadata
            Genre id for each document
        train_dataset, eval_dataset: torch.Dataset
            instantiated custom dataset object. Note that the model implementation and default trainer
            (mangoes.modeling.training_utils.CoreferenceFineTuneTrainer) are set up to work with
            mangoes.modeling.training_utils.MangoesCoreferenceDataset datasets, so take care when sending custom
            dataset arguments.
        compute_metrics: function
            The function that will be used to compute metrics at evaluation. Must return a dictionary string to metric
            values. Used by the trainer, see https://huggingface.co/transformers/training.html#trainer for more info.
        trainer: Transformers.Trainer
            custom instantiated trainer to use
        training_args:
            keyword arguments for training. For complete list, see
            https://huggingface.co/transformers/main_classes/trainer.html#transformers.TrainingArguments
        """
        if freeze_base:
            freeze_base_layers(self.model)
        if not trainer:
            if not train_dataset:
                if not train_documents or not train_cluster_ids or \
                        (not train_speaker_ids and self.model.config.task_specific_params["use_metadata"]) or\
                        (not train_genres and self.model.config.task_specific_params["use_metadata"]):
                    raise RuntimeError("Incomplete training data provided to train method")
                train_dataset = MangoesCoreferenceDataset(self.tokenizer,
                                                          self.model.config.task_specific_params["use_metadata"],
                                                          max_segment_len, max_segments, train_documents,
                                                          train_cluster_ids, train_speaker_ids, train_genres,
                                                          self.model.genres)
            if eval_documents and eval_cluster_ids and \
                    (self.model.config.task_specific_params["use_metadata"] or (eval_speaker_ids and eval_genres)) \
                    and not eval_dataset:
                eval_dataset = MangoesCoreferenceDataset(self.tokenizer,
                                                         self.model.config.task_specific_params["use_metadata"],
                                                         max_segment_len, max_segments, eval_documents,
                                                         eval_cluster_ids, eval_speaker_ids, eval_genres,
                                                         self.model.genres)
            if eval_dataset is not None and "evaluation_strategy" not in training_args:
                training_args["evaluation_strategy"] = "epoch"
            train_args = transformers.TrainingArguments(output_dir=output_dir, per_device_eval_batch_size=1,
                                                        per_device_train_batch_size=1,
                                                        label_names=["gold_starts", "gold_ends", "cluster_ids"],
                                                        **training_args)
            if not task_learn_rate:
                task_learn_rate = train_args.learning_rate
            trainer = CoreferenceFineTuneTrainer(task_learn_rate, "bert", self.model, args=train_args,
                                                 train_dataset=train_dataset, eval_dataset=eval_dataset,
                                                 tokenizer=self.tokenizer, compute_metrics=compute_metrics)
        trainer.train()
        self.model.eval()



# -*- coding: utf-8 -*-
import functools
import inspect
import time

import mangoes.utils.persist


def timer(display=print, label=None):
    """Decorator to measure the execution time of a function

    Parameters
    ----------
    display: callable
        function to use to display the execution time. Defaut = print but a logger should be used
    label
        label to give to the function in the display. If None (default), the name of the function will be used.

    """
    def decorated(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if label:
                label_to_display = label
            else:
                label_to_display = func.__name__

            display("Starting {} ...".format(label_to_display))
            start = time.time()
            response = func(*args, **kwargs)
            display("Finished {} in {}s".format(label_to_display, time.time() - start))
            return response
        return wrapper
    return decorated


def counter_filter(func):
    """Decorator to transform a function on a collections.Counter into a filter callable on this collections.Counter

    Filters are meant to be applied in the :func:`mangoes.corpus.Corpus.create_vocabulary` method to create a Vocabulary
    from a Corpus.
    The Counter object has to be the last parameter of the decorated function.
    So, this function, called without `words_count` parameter, provides a parametrized
    filter to be called on a collections.Counter.


    Examples
    ---------

    These 2 lines are equivalent :

        >>> mangoes.vocabulary.remove_most_frequent(min_frequency, words_count)
        >>> mangoes.vocabulary.remove_most_frequent(min_frequency)(words_count)


    This parametrized function is meant to be used to create a Vocabulary from a Corpus :

        >>> vocabulary = mangoes.Vocabulary(corpus,
        >>>                                filters=[mangoes.vocabulary.remove_most_frequent(max_frequency)])


    See Also
    --------
    :func:`mangoes.corpus.Corpus.create_vocabulary`
    """
    @functools.wraps(func)
    def _inner_filter(*args, **kwargs):
        try:
            if len(args) < len(inspect.signature(func).parameters):
                return functools.partial(func, *args, **kwargs)
            return func(*args, **kwargs)
        except mangoes.utils.exceptions.NotAllowedValue as e:
            raise e
    return _inner_filter



# -*- coding: utf-8 -*-
"""Classes and functions to evaluate embeddings.

"""

import warnings
warnings.warn("The evaluate module is deprecated and will be removed in next release: use mangoes.evaluation package "
              "instead", DeprecationWarning)

import abc
import heapq
import logging
from collections import namedtuple

import numpy as np
import scipy.stats

import mangoes.dataset
import mangoes.utils.arrays
import mangoes.utils.metrics

NA = "Not Applicable"  # TODO : temporary, find a better way to handle missing values


class _Result:
    """Abstract base class to store the result of evaluation tasks

    Warnings
    --------

    This class should not be used directly.
    Use derived classes instead.

    """
    LINE_LENGTH = 96
    COL = 4

    def __init__(self, dataset, predictions, oov=NA, ignored=NA):
        self._logger = logging.getLogger("{}.{}".format(__name__, self.__class__.__name__))

        self.dataset = dataset
        self.predictions = predictions

        self.score = self.get_score()
        self.oov = oov
        self.ignored = ignored

        self._results_dict = self._detail(self.dataset.subsets_to_questions)

    @property
    def summary(self):
        """Returns a printable string with a summary of the results of the evaluation

        Display the global score, the number of evaluated questions and the total number of questions in the dataset.

        See Also
        --------
        detail
        more_detail
        """
        return self.to_string(show_subsets=False)

    @property
    def detail(self):
        """Returns a printable string with the results of the evaluation for each subset of the dataset

        Display the score, the number of evaluated questions and the total number of questions in the subset
        for each subset of the dataset.

        See Also
        --------
        summary
        more_detail
        """
        return self.to_string(show_subsets=True, show_NA=False)

    @property
    def more_detail(self):
        """Returns a printable string with the detail of the results of the evaluation for each subset

        Display the score, the number of evaluated questions and the total number of questions and the prediction
        for each question and each subset of the dataset.

        See Also
        --------
        summary
        detail
        """
        return self.to_string(show_subsets=True, show_questions=True)

    def get_score(self, subset=None):
        """Returns the score of a subset of the dataset"""
        # TODO: implements macro/micro-average
        return self._compute_score([(g, self.predictions[q])
                                    for q, g in self.dataset.get_questions_and_gold(subset)
                                    if q in self.predictions])

    def _detail(self, data, name=""):
        result = {}

        for subset_name, subset_data in data.items():
            fullname = name + "/" + subset_name
            result[subset_name] = {}
            result[subset_name]["score"] = self.get_score(fullname)
            result[subset_name]["nb_questions_in_subset"] = mangoes.dataset.nb_questions(subset_data)
            if isinstance(subset_data, dict):
                result[subset_name]["subsets"] = self._detail(subset_data, name=fullname)
            else:
                questions = {}
                for question_w_gold in subset_data:
                    question = self.dataset.parse_question(question_w_gold).question
                    if question in self.predictions:
                        questions[question_w_gold] = self.predictions[question]
                if questions:
                    result[subset_name]["questions"] = questions
        return result

    def to_string(self, show_subsets=True, show_questions=False, show_NA=True):
        """Returns a printable version of the results

        Parameters
        ----------
        show_subsets: boolean
            if True, include the results of each subset of the dataset. If False, only the global score
        show_questions: boolean
            if True, show the predictions for each evaluated question
        show_NA: boolean
            if False, hide the subsets for which no question has been evaluated

        Returns
        -------
        str

        """
        string = ""
        for subset in self._results_dict.items():
            string += self._subset_to_string(subset, show_subsets=show_subsets, show_questions=show_questions,
                                             show_NA=show_NA, header=True)
        return string

    def _subset_to_string(self, subset, show_subsets=True, show_questions=False, show_NA=True,
                          header=False, dec=0):
        tab = " " * self.COL

        string = ""
        name, detail = subset

        if detail["score"]:
            if header:
                string += self._detail_header(dec, tab)
            else:
                string += "-" * self.LINE_LENGTH + "\n"
            string += self._detail_summary(name, detail, dec, tab)
        elif show_NA:
            string += self._detail_na(name, detail, dec, tab)

        if show_subsets and "subsets" in detail:
            for subset in detail["subsets"].items():
                string += self._subset_to_string(subset, dec=dec + 1, show_questions=show_questions, show_NA=show_NA)
        elif show_questions and "questions" in detail:
            string += "\n"
            string += self._detail_questions_header(dec, tab)
            for question, prediction in detail["questions"].items():
                string += self._detail_question(self.dataset.parse_question(question), prediction, dec, tab)
        return string

    @classmethod
    def _detail_na(cls, name, detail, dec, tab):
        line = "-" * cls.LINE_LENGTH + "\n"
        line += "{:<{width}}".format(tab * dec + name, width=(cls.LINE_LENGTH - 9 * cls.COL))
        line += "{:>{width}}".format("0/{}".format(detail["nb_questions_in_subset"]), width=3 * cls.COL)
        line += "{:>{width}}".format("Not available", width=6 * cls.COL)
        return line + "\n"

    @classmethod
    def _nb_questions(cls, subset):
        if isinstance(subset, list):
            return len(subset)
        if "questions" in subset:
            return len(subset["questions"])
        if "subsets" in subset:
            return sum([cls._nb_questions(subset["subsets"][s]) for s in subset["subsets"]])
        return 0

    @abc.abstractclassmethod
    def _compute_score(cls, data):
        pass

    @abc.abstractclassmethod
    def _detail_question(cls, question, prediction, dec, tab):
        pass

    @abc.abstractclassmethod
    def _detail_questions_header(cls, dec, tab):
        pass

    @abc.abstractclassmethod
    def _detail_summary(cls, name, score, dec, tab):
        pass

    @abc.abstractclassmethod
    def _detail_header(cls, dec, tab):
        pass


# ################ Analogy
class AnalogyResult(_Result):
    """Class to handle the result of analogies evaluation

    Attributes
    ----------
    dataset: mangoes.dataset.Dataset
        the dataset used to evaluate

    predictions: dict
        A dictionary where keys are the questions of the dataset (without gold) and the values are a tuple with
        the answers predicted with an Embedding using 3COSMUL and 3COSADD respectively

    score: AnalogyResult.Score
        Global score. The score is a tuple with the percentage of questions where the expected answer is in
        the predictions computed using, respectively, 3COSADD and 3COSMUL

    oov: set of strings
        words of the dataset that are not represented in the embedding

    ignored: int
        number of questions of the dataset ignored in the evaluation

    summmary
    detail
    more_detail

    """
    Score = namedtuple("Score", ["cosadd", "cosmul"])

    @classmethod
    def _compute_score(cls, data):
        if data:
            score_add, score_mul = 0, 0
            for gold, prediction in data:
                if gold in prediction[0]:
                    score_add += 1
                if gold in prediction[1]:
                    score_mul += 1
            return cls.Score(score_add / len(data), score_mul / len(data))

    @classmethod
    def _detail_header(cls, dec, tab):
        line = "{:>{width}}".format("", width=cls.LINE_LENGTH - 9 * cls.COL)
        line += "{:>{width}}".format("Nb questions", width=3 * cls.COL)
        line += "{:>{width}}".format("cosadd", width=3 * cls.COL)
        line += "{:>{width}}".format("cosmul", width=3 * cls.COL)
        return line + "\n" + "=" * cls.LINE_LENGTH + "\n"

    @classmethod
    def _detail_summary(cls, name, detail, dec, tab):
        score = detail["score"]
        nb_questions_in_subset = detail["nb_questions_in_subset"]
        nb_questions = cls._nb_questions(detail)

        line = "{:<{width}}".format(tab * dec + name, width=(cls.LINE_LENGTH - 3 * 3 * cls.COL))
        line += "{:>{width}}".format("{}/{}".format(nb_questions, nb_questions_in_subset), width=3 * cls.COL)
        line += "{:>{width}}".format("{:.2%}".format(score.cosadd), width=3 * cls.COL)
        line += "{:>{width}}".format("{:.2%}".format(score.cosmul), width=3 * cls.COL)

        return line + "\n"

    @classmethod
    def _detail_question(cls, question, prediction, dec, tab):
        question = " ".join(question)
        cosadd, cosmul = prediction

        line = "{:{width}}".format(tab * dec + question, width=(cls.LINE_LENGTH - 6 * cls.COL))
        line += "{:>{width}}".format(", ".join(cosadd), width=3 * cls.COL)
        line += "{:>{width}}".format(", ".join(cosmul), width=3 * cls.COL)
        return line + "\n"

    @classmethod
    def _detail_questions_header(cls, dec, tab):
        return ""


def analogy(embedding, dataset="all", allowed_answers=1, epsilon=0.001):
    """Evaluate an embedding on analogy task

    Parameters
    ----------
    embedding: mangoes.Embeddings
        The representation to evaluate
    dataset: mangoes.Dataset
        The dataset to use
    allowed_answers
        Number of answers
    epsilon: float
        Value used to prevent division by zero in 3CosMul

    Returns
    -------
    AnalogyResult

    """
    warnings.warn(
        "The evaluate module is deprecated and will be removed in next release: use mangoes.evaluation.analogy",
        DeprecationWarning)
    if not embedding or not dataset:
        return NA

    if dataset == "all":
        dataset = mangoes.dataset.load("analogy", language=embedding.words.language)

    selected_questions, common_words, oov, nb_ignored_questions = _filter_analogies(dataset, embedding)

    normalized_embedding = embedding.matrix.normalize()

    dataset_vocabulary = mangoes.Vocabulary(list(common_words))
    dataset_matrix = normalized_embedding[[embedding.words.index(word) for word in dataset_vocabulary]]

    if normalized_embedding.all_positive():
        similarities = mangoes.utils.metrics.pairwise_cosine_similarity(dataset_matrix,
                                                                        normalized_embedding,
                                                                        normalize=False)
    else:
        similarities = mangoes.utils.metrics.pairwise_non_negative_cosine_similarity(dataset_matrix,
                                                                                     normalized_embedding,
                                                                                     normalize=False)

    predictions = _evaluate_analogies(selected_questions, dataset_vocabulary, embedding.words, similarities,
                                      allowed_answers, epsilon)

    return AnalogyResult(dataset,
                         {question: prediction for question, prediction in zip(selected_questions, predictions)},
                         oov=oov, ignored=nb_ignored_questions)


def _filter_analogies(dataset, embedding):
    """From a dataset, keep only the questions where all words are in the dataset

    Parameters
    ----------
    dataset: mangoes.dataset.Dataset
    embedding: mangoes.Embeddings

    Returns
    -------
    questions : list of questions (question = list of string, without gold)
        list of the questions of the dataset where all words are in the dataset
    words : set of strings
        set of all the words of the questions
    oov_words: set of strings
        set of all the words from the dataset that are not represented in the embedding
    nb_ignored: int
        number of ignored questions
    """
    questions, words, oov_words = set(), set(), set()
    nb_ignored = 0

    for question_w_gold in dataset.questions_to_subsets:
        oov_words_in_question = [w for w in question_w_gold.split() if w not in embedding.words]
        if not oov_words_in_question:
            question, _ = question_w_gold.rsplit(maxsplit=1)
            questions.add(question)
            words.update(question.split())
        else:
            oov_words.update(oov_words_in_question)
            nb_ignored += 1

    return list(questions), words, oov_words, nb_ignored


def _evaluate_analogies(list_of_questions, dataset_vocabulary, candidate_vocabulary, similarities,
                        allowed_answers, epsilon):
    return [_resolve_analogy(question, dataset_vocabulary, candidate_vocabulary, similarities, allowed_answers, epsilon)
            for question in list_of_questions]


def _resolve_analogy(question, dataset_vocabulary, candidate_vocabulary, similarities,
                     allowed_answers, epsilon):
    a, b, c = question.split()

    a_similarities = similarities[dataset_vocabulary.index(a)]
    b_similarities = similarities[dataset_vocabulary.index(b)]
    c_similarities = similarities[dataset_vocabulary.index(c)]

    question_terms_indices = [candidate_vocabulary.index(word) for word in (a, b, c)]

    scores_add = b_similarities - a_similarities + c_similarities
    scores_add[question_terms_indices] = np.NINF
    best_answer_add = [candidate_vocabulary[i] for i in _get_n_best(allowed_answers, scores_add)]

    scores_mul = b_similarities * c_similarities / (a_similarities + epsilon)
    scores_mul[question_terms_indices] = np.NINF
    best_answer_mul = [candidate_vocabulary[i] for i in _get_n_best(allowed_answers, scores_mul)]

    return best_answer_add, best_answer_mul


def _get_n_best(nb_best, scores):
    if nb_best == 1:
        return [np.nanargmax(scores)]

    return heapq.nlargest(nb_best, range(len(scores)), key=lambda i: scores[i])


# ############## Similarity

class SimilarityResult(_Result):
    """Class to handle the result of word similarity evaluation

    Attributes
    ----------
    dataset: mangoes.dataset.Dataset
        the dataset used to evaluate

    predictions: dict
        A dictionary where keys are the questions of the dataset (without gold) and the values are the computed
        similarity

    score: SimilarityResult.Score
        Global score. The score is a tuple of the 2 correlation coefficient : Pearson and Spearman, respectively.
        Each coefficient is a tuple with the coefficient itself and the p-value.

    oov: set of strings
        words of the dataset that are not represented in the embedding

    ignored: int
        number of questions of the dataset ignored in the evaluation

    summmary
    detail
    more_detail

    """
    Coeff = namedtuple("Coeff", ["coeff", "pvalue"])
    Score = namedtuple("Score", ["pearson", "spearman"])

    @classmethod
    def _compute_score(cls, data):
        if data:
            gold = [float(d[0]) for d in data]
            predictions = [d[1] for d in data]

            pearson = cls.Coeff(*scipy.stats.pearsonr(predictions, gold))
            spearman = cls.Coeff(*scipy.stats.spearmanr(predictions, gold))

            return cls.Score(pearson, spearman)

    @classmethod
    def _detail_header(cls, dec, tab):
        line = "{:>{width}}".format("", width=cls.LINE_LENGTH - 8 * cls.COL)
        line += "{:>{width}}".format("pearson", width=4 * cls.COL)
        line += "{:>{width}}".format("spearman", width=4 * cls.COL)
        line += "\n"
        line += "{:>{width}}".format("", width=cls.LINE_LENGTH - 11 * cls.COL)
        line += "{:>{width}}".format("Nb questions", width=3 * cls.COL)
        line += "{:>{width}}".format("(p-value)", width=4 * cls.COL)
        line += "{:>{width}}".format("(p-value)", width=4 * cls.COL)
        return line + "\n" + "=" * cls.LINE_LENGTH + "\n"

    @classmethod
    def _detail_summary(cls, name, detail, dec, tab):
        score = detail["score"]
        nb_questions_in_subset = detail["nb_questions_in_subset"]
        nb_questions = cls._nb_questions(detail)

        line = "{:<{width}}".format(tab * dec + name, width=(cls.LINE_LENGTH - 11 * cls.COL))
        line += "{:>{width}}".format("{}/{}".format(nb_questions, nb_questions_in_subset), width=3 * cls.COL)
        line += "{:>{width}}".format("{:.2}({:.2e})".format(*score.pearson), width=4 * cls.COL)
        line += "{:>{width}}".format("{:.2}({:.2e})".format(*score.spearman), width=4 * cls.COL)
        return line + "\n"

    @classmethod
    def _detail_questions_header(cls, dec, tab):
        width = (cls.LINE_LENGTH - dec * cls.COL) // 4

        line = tab * dec
        line += "{:{width}}".format("", width=width)
        line += "{:{width}}".format("", width=width)
        line += "{:{width}}".format("gold", width=width)
        line += "{:{width}}".format("score", width=width)
        return line + "\n"

    @classmethod
    def _detail_question(cls, question, prediction, dec, tab):
        width = (cls.LINE_LENGTH - dec * cls.COL) // 4
        question, gold = question
        word1, word2 = question.split()

        line = tab * dec
        line += "{:{width}}".format(word1, width=width)
        line += "{:{width}}".format(word2, width=width)
        line += "{:{width}}".format(gold, width=width)
        line += "{:{width}}".format("{:.2f}".format(prediction), width=width)
        return line + "\n"


def similarity(embedding, dataset="all", metric=mangoes.utils.metrics.rowwise_cosine_similarity):
    """Evaluate an embedding on word similarity task

    Parameters
    ----------
    embedding: mangoes.Embeddings
        The representation to evaluate
    dataset: mangoes.Dataset
        The dataset to use
    metric
        the metric to use to compute the similarity (default : cosine)

    Returns
    -------
    SimilarityResult

    """
    warnings.warn(
        "The evaluate module is deprecated and will be removed in next release: use mangoes.evaluation.similarity",
        DeprecationWarning)
    if not embedding or not dataset:
        return NA

    if dataset == "all":
        dataset = mangoes.dataset.load("similarity", language=embedding.words.language)

    selected_questions, _, oov, nb_ignored_questions = _filter_similarities(dataset, embedding)

    predictions = _evaluate_similarities(selected_questions, embedding, metric)

    return SimilarityResult(dataset,
                            {question: prediction for question, prediction in zip(selected_questions, predictions)},
                            oov=oov, ignored=nb_ignored_questions)


def _evaluate_similarities(words_pairs, embedding, metric):
    words_pairs_indices = np.array([[embedding.words.index(q.split()[0]),
                                     embedding.words.index(q.split()[1])]
                                    for q in words_pairs])

    first_terms = embedding.matrix[words_pairs_indices[:, 0], :]
    second_terms = embedding.matrix[words_pairs_indices[:, 1], :]

    predictions = metric(first_terms, second_terms)

    return predictions


def _filter_similarities(dataset, embedding):
    """From a dataset, keep only the questions where all words are in the dataset

    Parameters
    ----------
    dataset: mangoes.dataset.Dataset
    embedding: mangoes.Embeddings

    Returns
    -------
    questions : list of questions (question = list of string, without gold)
        list of the questions of the dataset where all words are in the dataset
    words : set of strings
        set of all the words of the questions
    oov_words: set of strings
        set of all the words from the dataset that are not represented in the embedding
    nb_ignored: int
        number of ignored questions
    """
    questions, words, oov_words = set(), set(), set()
    nb_ignored = 0

    for question_w_gold in dataset.questions_to_subsets:
        oov_words_in_question = [w for w in dataset.parse_question(question_w_gold).question.split()
                                 if w not in embedding.words]
        if not oov_words_in_question:
            question = dataset.parse_question(question_w_gold).question
            questions.add(question)
            words.update(question.split())
        else:
            oov_words.update(oov_words_in_question)
            nb_ignored += 1

    return list(questions), words, oov_words, nb_ignored


# ############## Outlier Detection

class OutlierDetectionResult(_Result):
    """Class to handle the result of outlier detection evaluation

    Attributes
    ----------
    dataset: mangoes.dataset.Dataset
        the dataset used to evaluate

    predictions: dict
        A dictionary where keys are the questions of the dataset and the values are the words of the question, sorted
        according to their compactness score

    score: OutlierDetectionResult.Score
        Global score. The score is a tuple with the Outlier Position Percentage and the Accuracy measures.

    oov: set of strings
        words of the dataset that are not represented in the embedding

    ignored: int
        number of questions of the dataset ignored in the evaluation

    summmary
    detail
    more_detail

    """
    Score = namedtuple("Score", ["opp", "accuracy"])

    def _detail(self, data, name=""):
        result = {}

        for subset_name, subset_data in data.items():
            fullname = name + "/" + subset_name

            result[subset_name] = {}
            result[subset_name]["score"] = self.get_score(fullname)
            result[subset_name]["nb_questions_in_subset"] = mangoes.dataset.nb_questions(subset_data)

            if isinstance(subset_data, dict):
                result[subset_name]["subsets"] = self._detail(subset_data, name=fullname)
            else:
                questions = {}
                for question_w_gold in subset_data:
                    if question_w_gold in self.predictions:
                        questions[question_w_gold] = self.predictions[question_w_gold]
                if questions:
                    result[subset_name]["questions"] = questions
        return result

    @classmethod
    def _detail_header(cls, dec, tab):
        line = "{:>{width}}".format("", width=cls.LINE_LENGTH - 9 * cls.COL)
        line += "{:>{width}}".format("Nb questions", width=3 * cls.COL)
        line += "{:>{width}}".format("OPP", width=3 * cls.COL)
        line += "{:>{width}}".format("accuracy", width=3 * cls.COL)
        return line + "\n" + "=" * cls.LINE_LENGTH + "\n"

    @classmethod
    def _detail_summary(cls, name, detail, dec, tab):
        score = detail["score"]
        nb_questions_in_subset = detail["nb_questions_in_subset"]
        nb_questions = cls._nb_questions(detail)

        line = "{:<{width}}".format(tab * dec + name, width=(cls.LINE_LENGTH - 9 * cls.COL))
        line += "{:>{width}}".format("{}/{}".format(nb_questions, nb_questions_in_subset), width=3 * cls.COL)
        line += "{:>{width}}".format("{:.2%}".format(score.opp), width=3 * cls.COL)
        line += "{:>{width}}".format("{:.2%}".format(score.accuracy), width=3 * cls.COL)
        return line + "\n"

    @classmethod
    def _detail_questions_header(cls, dec, tab):
        line = "{:>{width}}".format("outlier position", width=cls.LINE_LENGTH)
        return line + "\n"

    @classmethod
    def _detail_question(cls, question, prediction, dec, tab):
        line = "{:{width}}".format(tab * dec + question.question, width=(cls.LINE_LENGTH - 2 * cls.COL))
        line += "{:>{width}}".format(prediction.index(question.gold), width=2 * cls.COL)
        return line + "\n"

    @classmethod
    def _compute_score(cls, data):
        if data:
            gold = [d[0] for d in data]
            predictions = [d[1] for d in data]

            outlier_positions = [p.index(g) + 1 for p, g in zip(predictions, gold)]
            outlier_detections = [op == len(p) for p, op in zip(predictions, outlier_positions)]

            opp_score = sum([op / len(p) for p, op in zip(predictions, outlier_positions)]) / len(predictions)
            accuracy = sum(outlier_detections) / len(predictions)

            return cls.Score(opp_score, accuracy)


def outlier_detection(embedding, dataset="all"):
    """Evaluate an embedding on outlier detection task

     Parameters
     ----------
     embedding: mangoes.Embeddings
         The representation to evaluate
     dataset: mangoes.Dataset
         The dataset to use

     Returns
     -------
     SimilarityResult

     """
    warnings.warn(
        "The evaluate module is deprecated and will be removed in next release: use mangoes.evaluation.outlier",
        DeprecationWarning)
    if not embedding or not dataset:
        return NA

    if dataset == "all":
        dataset = mangoes.dataset.load("outlier_detection", language=embedding.words.language)

    selected_questions, _, oov, nb_ignored_questions = _filter_clusters(dataset, embedding)

    predictions = _evaluate_outliers(selected_questions, embedding)

    return OutlierDetectionResult(dataset,
                                  {question: prediction for question, prediction in
                                   zip(selected_questions, predictions)},
                                  oov=oov, ignored=nb_ignored_questions)


def _filter_clusters(dataset, embedding):
    questions, words, oov_words = set(), set(), set()
    nb_ignored = 0

    for question_w_gold, subset_path in dataset.questions_to_subsets.items():
        oov_words_in_question = [w for w in question_w_gold.split() if w not in embedding.words]
        if not oov_words_in_question:
            questions.add(question_w_gold)
            words.update(question_w_gold.split())
        else:
            oov_words.update(oov_words_in_question)
            nb_ignored += 1

    return list(questions), words, oov_words, nb_ignored


def _evaluate_outliers(selected_clusters, embedding):
    result = []
    for cluster in selected_clusters:
        compactness_scores = []
        for w in cluster.split():
            all_except_w = mangoes.utils.arrays.Matrix.factory(embedding.matrix[[embedding.words.index(word)
                                                                                 for word in cluster.split()
                                                                                 if word is not w]])
            compactness_scores.append(_pseudo_inversed_compactness_score(all_except_w, embedding[w]))

        sorted_indices = reversed(np.asarray(compactness_scores, dtype=float).argsort())
        result.append([cluster.split()[i] for i in sorted_indices])
    return result


def _pseudo_inversed_compactness_score(vectors, w):
    similarities = mangoes.utils.metrics.pairwise_cosine_similarity(w, vectors)
    return similarities.sum()


import mangoes.evaluation.statistics


# ########################################
# Partition function
def isotropy_from_partition_function(*args, **kwargs):
    warnings.warn(
        "The evaluate module is deprecated and will be removed in next release: use mangoes.evaluation.statistics",
        DeprecationWarning)

    return mangoes.evaluation.statistics.isotropy_from_partition_function(*args, **kwargs)


# ########################################
# Distances between words

def distances_one_word_histogram(*args, **kwargs):
    warnings.warn(
        "The evaluate module is deprecated and will be removed in next release: use mangoes.evaluation.statistics",
        DeprecationWarning)
    return mangoes.evaluation.statistics.distances_one_word_histogram(*args, **kwargs)


def distances_histogram(*args, **kwargs):
    warnings.warn(
        "The evaluate module is deprecated and will be removed in next release: use mangoes.evaluation.statistics",
        DeprecationWarning)
    return mangoes.evaluation.statistics.distances_histogram(*args, **kwargs)


# ########################################
# t-SNE
def tsne(embeddings):
    """Create a 2d projections of the embeddings using t-SNE

    Parameters
    ----------
    embeddings: an instance of Embeddings
        Instance of :class:`mangoes.Embeddings` to project
    """
    warnings.warn(
        "The evaluate module is deprecated and will be removed in next release: use mangoes.visualize module instead",
        DeprecationWarning)

    try:
        matrix = embeddings.matrix.toarray()
    except AttributeError:
        # already dense
        matrix = embeddings.matrix

    import sklearn.manifold
    model = sklearn.manifold.TSNE()
    return model.fit_transform(matrix)



# -*- coding: utf-8 -*-
"""Dimensionality reduction to apply the co-occurrence count matrix.

This module provides reductions that can be used in the `transformations` parameter of
the :func:`mangoes.create_representation` function to create an Embeddings from a CooccurrenceCount.

Examples
---------

import mangoes.base    >>> pca = mangoes.reduction.PCA(dimensions=50)
    >>> embeddings = mangoes.base.create_representation(cc, transformations=pca)


See Also
--------
:func:`mangoes.create_representation`
:class:`mangoes.Transformation`
"""
import logging

import numpy as np
import scipy.sparse
import scipy.sparse.linalg
import sklearn.decomposition

import mangoes.base
import mangoes.utils.exceptions

_logger = logging.getLogger(__name__)


class PCA(mangoes.base.Transformation):
    """Defines a transformation that apply PCA

    Parameters
    ----------
    dimensions: int
        desired dimensionality of the returned matrix

    """
    def __init__(self, dimensions):
        super().__init__()
        self._params["dimensions"] = dimensions

    @property
    def dimensions(self):
        return self._params["dimensions"]

    def __call__(self, matrix):
        """Performs the PCA on the matrix and returns representations for target words

        Given a count matrix, returns a matrix reduced to given dimension using PCA

        Parameters
        ----------
        matrix: a matrix, sparse or dense

        Returns
        -------
        matrix
        """
        # TODO : check performances - power iteration
        # TODO : or alternative (center first ?, optional)
        transformer = sklearn.decomposition.TruncatedSVD(n_components=self.dimensions)
        return transformer.fit_transform(matrix)


class SVD(mangoes.base.Transformation):
    """Defines a transformation to reduce dimensionality using SVD

    Given a count matrix :math:`M`, the SVD decomposition gives :

    .. math::
        M_d = U_d.\Sigma_d.V_d^\\top

    by keeping the top :math:`d` eigenvalues in :math:`\Sigma` where :math:`d` = dimensions

    **If parameter `add_context_vectors` is False (default) :**

    We get :math:`W_d`, a matrix representing the original target words (rows of the matrix) reduced to
    given dimensions :

    .. math::
        W_d = U_d.\Sigma_d^{weight}

    The function will return :math:`W_d`


    **If parameter `add_context_vectors` is True :**

    If the same vocabulary as been used as rows and columns to construct the matrix, you can also set
    `add_context_vectors` to True to add the "context vectors" :math:`C_d` to :math:`W_d`.

    The `symmetric` parameter define the way to construct :math:`C_d` :

    If symmetric = False :

    .. math::
        C_d = V_d.\Sigma_d^{1-weight}

    So weight = 1 corresponds to the traditional SVD factorization :math:`W_d = U_d.\Sigma_d, C_d = V_d`

    If symmetric = True :

    .. math::
        C_d = V_d.\Sigma_d^{weight}

    So with weight = 0 : :math:`W_d = U_d, C_d = V_d`

    The function will return :math:`W_d + C_d`


    Notes
    -----
    `dimensions` has to be lower than both dimensions of the input matrix

    Warnings
    --------

    * `add_context_vectors` should be used **only if** represented words and words used as contexts are the same
    * weight should be a value between 0 and 1

    Attributes
    ----------
    dimensions: int
        desired dimensionality of the returned matrix. Must be less than both dimensions of the original.
    weight: {1, 0, 0.5}
        a parameter that defines the way to compute the matrix (see above)
    add_context_vectors: boolean
        Use the context vectors in addition to the words vectors (should be used **only if** represented words and
        words used as contexts are the same). Default = False.
    symmetric: boolean
        if True, the true matrices :math:`W_d` and :math:`C_d` will be built symmetrically (see above)

    """
    def __init__(self, dimensions, weight=1, add_context_vectors=False, symmetric=False):
        super().__init__()
        self._params["dimensions"] = dimensions
        self._params["weight"] = weight
        self._params["add_context_vectors"] = add_context_vectors
        self._params["symmetric"] = symmetric

    @property
    def dimensions(self):
        return self._params["dimensions"]

    @property
    def weight(self):
        return self._params["weight"]

    @property
    def add_context_vectors(self):
        return self._params["add_context_vectors"]

    @property
    def symmetric(self):
        return self._params["symmetric"]

    def __call__(self, matrix):
        """Performs the reduction

        Parameters
        ----------
        matrix: a matrix, sparse or dense

        Returns
        -------
        matrix
        """
        words_matrix, contexts_matrix = _svd(matrix, self.dimensions, symmetric=self.symmetric, weight=self.weight)

        if self.add_context_vectors:
            return words_matrix + contexts_matrix
        return words_matrix


def _svd(matrix, dimensions, symmetric=False, weight=1):
    """Performs SVD on the matrix and returns representations for target words and context words

    See mangoes.reduction.svd

    Notes
    ------
    `dimensions` has to be lower than both dimensions

    Parameters
    ----------
    matrix: matrix
        can be sparse or dense
    dimensions:
        desired dimensionality of the returned matrix. Must be less than both dimensions of the original.
    symmetric: boolean
        if True, the both matrices Wd and Cd will be built symmetrically (see above)
    weight: {1, 0, 0.5}
        a parameter that defines the way to compute the matrix (see above)

    Returns
    -------
    tuple
        the two matrices : (words_matrix, context_words_matrix)
    """
    # TODO : check performances - power iteration
    if dimensions > min(matrix.shape):
        msg = "'dimensions' parameter must be between 1 and min(matrix.shape). " \
              "dimensions={}, matrix.shape={}".format(dimensions, matrix.shape)
        raise mangoes.utils.exceptions.IncompatibleValue(msg)

    if weight < 0 or weight > 1:
        _logger.warning("weight value should be between 0 and 1 (weight={})".format(weight))

    matrix = matrix.astype(float)
    Ud, sd, Vdt = scipy.sparse.linalg.svds(matrix, k=dimensions)
    Vd = Vdt.T

    if symmetric and weight == 0:
        # no need to construct Sigma
        return Ud, Vd

    Sd_p = np.zeros((dimensions, dimensions))
    Sd_p[:dimensions, :dimensions] = np.diag(sd**weight)

    if symmetric:
        return Ud.dot(Sd_p), Vd.dot(Sd_p)

    Sd_1_p = np.zeros((dimensions, dimensions))
    Sd_1_p[:dimensions, :dimensions] = np.diag(sd**(1 - weight))
    return Ud.dot(Sd_p), Vd.dot(Sd_1_p)



# -*-coding:utf-8 -*
"""Set of Mangoes's Exceptions

"""
import logging

logger = logging.getLogger(__name__)


class MangoesError(Exception):
    """Basic exception for errors raised by mangoes"""

    def __init__(self, msg, original_exception=None):
        if original_exception:
            msg += " : " + original_exception
        super(MangoesError, self).__init__("{}".format(msg))
        self.original_exception = original_exception


class NotAllowedValue(MangoesError, ValueError):
    """Raise when a value is not allowed"""

    def __init__(self, value=None, allowed_values=None, msg=None):
        if msg is None:
            msg = ""
            if value:
                msg += "'{}' is not allowed. ".format(value)
            if allowed_values:
                msg += "Allowed values are {}".format(", ".join(["'{}'".format(v) for v in allowed_values]))
        super(NotAllowedValue, self).__init__(msg)


class IncompatibleValue(MangoesError, ValueError):
    """Raise when 2 values are not compatible"""


class UnsupportedType(MangoesError, TypeError):
    """Raise when the type of an argument is not supported"""


class ResourceNotFound(MangoesError, FileNotFoundError):
    """Raise when an expected resource is not found on a given path"""

    def __init__(self, path=None, msg=None):
        if msg is None:
            msg = "Resource '{}' does not exist".format(path)
        super(ResourceNotFound, self).__init__(msg)


class RequiredValue(MangoesError):  # TODO : or missing value ?
    """Raise when a required argument is empty"""

class RuntimeError(MangoesError, RuntimeError):
    """Raised when an error is detected that doesn’t fall in any of the other categories.
    The associated value is a string indicating what precisely went wrong."""

class OutOfVocabulary(MangoesError, ValueError):
    """Raise when a word is not in a vocabulary"""

    def __init__(self, value=None, msg=None):
        if msg is None:
            msg = ""
            if value:
                msg += "'{}' is out of vocabulary. ".format(value)
        super(OutOfVocabulary, self).__init__(msg)



# -*- coding: utf-8 -*-
"""
This module provides helper classes for training huggingface models using the interface in mangoes.modeling.bert.
These classes are mainly called internally from mangoes.modeling.bert, however these classes can be instantiated on
their own (or subclassed) and passed to training methods in mangoes.modeling.bert for more customization/control.
"""
import fileinput
import collections
import random

import transformers
import torch
from torch.utils.data import IterableDataset, DataLoader, Dataset


def freeze_base_layers(model):
    """
    Function to freeze the base layer in a fine tuning/pretraining model
    """
    for param in model.base_model.parameters():
        param.requires_grad = False


class MangoesLineByLineIterableDataset(IterableDataset):
    """
    Subclass of torch.utils.data.IterableDataset.
    Used for large text datasets that are not to be loaded into memory at the same time.

    Parameters
    ----------
    filenames: str or List[str]
        paths to file containing text
    tokenizer: transformers.Tokenizer
    max_len: int
        size of input tensors. If None, will default to tokenizer.model_max_length
    length: int
        number of lines in the dataset. If None, will lazily calculate it
    encoding: str
        encoding of text files
    """
    def __init__(self, filenames, tokenizer, max_len=None, length=None, encoding=None):
        if isinstance(filenames, str):
            filenames = [filenames]
        self.filenames = filenames
        self.tokenizer = tokenizer
        self.block_size = max_len if max_len else tokenizer.model_max_length
        self.encoding = encoding
        self.length = length

    def preprocess(self, text):
        """
        Preprocess a line of text using the tokenizer

        Parameters
        ----------
        text: str

        Returns
        -------
        dictionary containing input_id encodings
        """
        batch_encoding = self.tokenizer(text.strip("\n"), truncation=True, max_length=self.block_size)
        examples = batch_encoding["input_ids"]
        return {"input_ids": torch.tensor(examples, dtype=torch.long)}

    def line_mapper(self, line):
        return self.preprocess(line)

    def __iter__(self):
        reader = fileinput.FileInput(self.filenames, mode="r",
                                     openhook=None if not self.encoding else fileinput.hook_encoded(self.encoding))
        mapped_itr = map(self.line_mapper, reader)
        return mapped_itr

    def __len__(self):
        if not self.length:
            self.length = self._get_length()
        return self.length

    def _get_length(self):
        """
        Helper function to get number of line in texts
        """
        return sum(1 for _ in fileinput.FileInput(self.filenames, mode="r",
                                                  openhook=None if not self.encoding else
                                                  fileinput.hook_encoded(self.encoding)))


class MangoesTextClassificationDataset(Dataset):
    """
    Subclass of torch.utils.data.Dataset class for sequence or token classification.
    To be used with transformers.Trainer.

    Parameters
    ----------
    texts: List[str]
    labels: List[str or int] if seq classification, else List[List[str or int]] if token classification
        Labels can be raw strings, which will us label2id to convert to ids, or the ids themselves.
    tokenizer: transformers.Tokenizer
    max_len: int
        max length of input sequences, if None, will default to tokenizer.model_max_length
    label2id: dict of str -> int
        if labels are not already converted to output ids, dictionary with mapping to use.
    """
    def __init__(self, texts, labels, tokenizer, max_len=None, label2id=None):
        self.texts = texts
        token_classes = True if isinstance(labels[0], list) else False
        if token_classes:
            # if token classification
            raw_labels = True if isinstance(labels[0][0], str) else False
        else:
            # if sequence classification
            raw_labels = True if isinstance(labels[0], str) else False
        if raw_labels and not label2id:
            raise TypeError('Labels passed to dataset are not converted into output ids and no id mapping was passed.')
        if raw_labels:
            if token_classes:
                labels = [[label2id[label] for label in sublist] for sublist in labels]
            else:
                labels = [label2id[label] for label in labels]
        self.labels = labels
        self.tokenizer = tokenizer
        self.max_len = max_len if max_len else tokenizer.model_max_length

    def __len__(self):
        return len(self.texts)

    def __getitem__(self, item):
        text = str(self.texts[item])
        labels = self.labels[item]
        encoding = self.tokenizer(
            text,
            add_special_tokens=True,
            max_length=self.max_len,
            return_token_type_ids=False,
            padding="max_length",
            return_attention_mask=True,
            return_tensors='pt',
            truncation=True
        )
        return {
            'input_ids': encoding['input_ids'].flatten(),
            'attention_mask': encoding['attention_mask'].flatten(),
            'labels': torch.tensor(labels, dtype=torch.long)
        }


class MangoesLineByLineDataset(Dataset):
    """
    Subclass of torch.utils.data.Dataset class for pretraining language models.
    Similar to transformers.LineByLineTextDataset, but supports multiple input files

    Parameters
    ----------
    filenames: str or List[str]
        paths to files to include in dataset.
    tokenizer: transformers.Tokenizer
    max_len: int
        max length of input sequences, if None, will default to tokenizer.model_max_length
    encoding: str
        encoding of text files
    """
    def __init__(self, filenames, tokenizer, max_len=None, encoding=None):
        if isinstance(filenames, str):
            filenames = [filenames]
        self.filenames = filenames
        self.encoding = encoding
        with fileinput.input(self.filenames, mode="r", openhook=None if not self.encoding else
                             fileinput.hook_encoded(self.encoding)) as f:
            lines = [line for line in f if (len(line) > 0 and not line.isspace())]

        batch_encoding = tokenizer(lines, add_special_tokens=True, truncation=True, max_length=max_len)
        self.examples = batch_encoding["input_ids"]
        self.examples = [{"input_ids": torch.tensor(example, dtype=torch.long)} for example in self.examples]

    def __len__(self):
        return len(self.examples)

    def __getitem__(self, i):
        return self.examples[i]


class MangoesQuestionAnsweringDataset(Dataset):
    """
    Subclass of Torch Dataset for question answering datasets. Currently meant to work with BERT models.

    Parameters
    ----------
    tokenizer: transformers.Tokenizer
    question_texts: List of str
        The texts corresponding to the questions
    context_texts: List of str
        The texts corresponding to the contexts
    answer_texts: List of str
        The texts corresponding to the answers
    start_indices: List of int
        The character positions of the start of the answers
    max_seq_length:int
        The maximum total input sequence length after tokenization.
    doc_stride: int
        When splitting up a long document into chunks, how much stride to take between chunks.
    max_query_length: int
        The maximum number of tokens for the question.
    """
    def __init__(self, tokenizer, question_texts, context_texts, answer_texts, start_indices, max_seq_length=384,
                 doc_stride=128, max_query_length=64):

        if isinstance(tokenizer, transformers.PreTrainedTokenizerFast):
            import os
            # TODO: (02/12/2020) currently the FastTokenizers don't work with QAPiplines, so we have to typecast
            # https://github.com/huggingface/transformers/issues/7735
            # once this is fixed in transformers, we should remove this hacky fix
            tokenizer.save_vocabulary("./")
            tokenizer = transformers.BertTokenizer("./vocab.txt", **tokenizer.init_kwargs)
            os.remove("./vocab.txt")
        # convert to squad examples
        if not len(question_texts) == len(answer_texts) or not len(question_texts) == len(start_indices) or \
                not len(question_texts) == len(context_texts):
            raise ValueError("Question Answering dataset needs answers, contexts, and start indices for every example")
        examples = []
        for i in range(len(question_texts)):
            examples.append(transformers.SquadExample(qas_id=len(examples),
                                                      question_text=question_texts[i],
                                                      context_text=context_texts[i],
                                                      answer_text=answer_texts[i],
                                                      start_position_character=start_indices[i],
                                                      title=""))
        self.features = transformers.squad_convert_examples_to_features(examples, tokenizer,
                                                                        max_seq_length=max_seq_length,
                                                                        doc_stride=doc_stride,
                                                                        max_query_length=max_query_length,
                                                                        is_training=True)

    def __len__(self):
        return len(self.features)

    def __getitem__(self, i):
        feature = self.features[i]

        inputs = {
            "input_ids": torch.tensor(feature.input_ids, dtype=torch.long),
            "attention_mask": torch.tensor(feature.attention_mask, dtype=torch.long),
            "token_type_ids": torch.tensor(feature.token_type_ids, dtype=torch.long),
        }
        start_positions = torch.tensor(feature.start_position, dtype=torch.long)
        end_positions = torch.tensor(feature.end_position, dtype=torch.long)
        inputs.update({"start_positions": start_positions, "end_positions": end_positions})
        return inputs


class MangoesCoreferenceDataset(Dataset):
    """
    Subclass of Torch Dataset for co-reference datasets such as Ontonotes. Currently meant to work with BERT models.

    Each example is one document. Documents are parsed by first tokenizing each sentence then aggregating sentences into
    segments, keeping track of label, metadata, and sentence indices.

    Parameters
    ----------
    tokenizer: transformers.BertTokenizerFast
        tokenizer to use
    use_metadata: Boolean
        Whether or not to use speaker ids and genres
    max_segment_len: int
        maximum number of sub-tokens for one segment
    max_segments: int
        Maximum number of segments to return per __getitem__ (ie per document)
    documents: List of Lists of Lists of strings
        Text for each document. As cluster ids are labeled by word, a document is a list of sentences. One
        sentence is a list of words (ie already split on whitespace/punctuation)
    cluster_ids: List of Lists of Lists of (ints or Tuple(int, int))
        Cluster ids for each word in documents argument. Assumes words that aren't mentions have either None or -1 as
        id. In the case where a word belongs to two different spans (with different cluster ids), the cluster id for
        word should be a tuple of ints corresponding to the different cluster ids.
    speaker_ids: List of Lists of Lists of ints
        Speaker id for each word in documents. Assumes positive ids (special tokens (such as [CLS] and [SEP] that are
        added at beginning and end of segments) will be assigned speaker ids of -1)
    genres: List of ints or strings
        Genre (id) for each document. If strings, genre_to_id parameter needs to not be None
    genre_to_id: dict of string->int
        Mapping of genres to their id number.
    """
    def __init__(self, tokenizer, use_metadata, max_segment_len, max_segments, documents, cluster_ids, speaker_ids=None,
                 genres=None, genre_to_id=None):
        self.use_metadata = use_metadata
        if (use_metadata and not speaker_ids) or (use_metadata and not genres):
            raise RuntimeError("use_metadata argument is set to True in MangoesCoreferenceDataset init function, but "
                               "missing speaker and/or genre input data")
        if use_metadata and isinstance(genres[0], str) and not genre_to_id:
            raise RuntimeError("Input genre data has not been converted to ids yet, and genre_to_id parameter is "
                               "unfilled")
        self.examples = []
        self.max_segments = max_segments
        for i in range(len(documents)):
            # for each sentence, tokenize into word pieces then aggregate input ids, clusters, speakers
            subtoken_ids = []
            subtoken_cluster_ids = []
            subtoken_speakers = []
            subtoken_offset_mappings = []
            for j in range(len(documents[i])):
                encoding = tokenizer(documents[i][j], add_special_tokens=False, is_split_into_words=True,
                                     return_offsets_mapping=True)
                subtoken_ids.append(encoding["input_ids"])
                subtoken_cluster_ids.append(self.get_subtoken_data(cluster_ids[i][j], encoding["offset_mapping"]))
                subtoken_offset_mappings.append(encoding["offset_mapping"])
                if use_metadata and speaker_ids:
                    subtoken_speakers.append(self.get_subtoken_data(speaker_ids[i][j], encoding["offset_mapping"]))

            # aggregate into segments
            assert len(subtoken_ids) == len(subtoken_cluster_ids)
            current_segment_ids = []
            current_segment_cluster_ids = []
            current_segment_speaker_ids = []
            current_sentence_map = []
            segments_ids = []
            segments_clusters = []
            segments_speakers = []
            segments_attention_mask = []
            sentence_map = []
            for j in range(len(subtoken_ids)):
                if len(current_segment_ids) + len(subtoken_ids[j]) <= max_segment_len - 2:
                    current_segment_ids += subtoken_ids[j]
                    current_segment_cluster_ids += subtoken_cluster_ids[j]
                    current_sentence_map += [j] * len(subtoken_ids[j])
                    if use_metadata and speaker_ids:
                        current_segment_speaker_ids += subtoken_speakers[j]
                else:
                    if len(current_segment_ids) > 0:
                        # segments contain cls and sep special tokens at beginning and end for BERT processing
                        segments_ids.append(self.pad_list([tokenizer.cls_token_id] + current_segment_ids +
                                                          [tokenizer.sep_token_id], max_segment_len,
                                                          tokenizer.convert_tokens_to_ids(tokenizer.pad_token)))
                        segments_clusters.append(self.pad_list([None] + current_segment_cluster_ids + [None],
                                                               max_segment_len, None))
                        segments_attention_mask.append(self.pad_list([1] * (len(current_segment_ids) + 2),
                                                                     max_segment_len))
                        sentence_map += [current_sentence_map[0]] + current_sentence_map + [current_sentence_map[-1]]
                        if use_metadata and speaker_ids:
                            segments_speakers.append(self.pad_list([-1] + current_segment_speaker_ids + [-1],
                                                                   max_segment_len))
                    if len(subtoken_ids[j]) > max_segment_len - 2:
                        # if sentence j is longer than max_seq_len, create segment out of as much as possible,
                        # then remove these from sentence j and continue
                        segment_stop_index = max_segment_len - 2
                        while subtoken_offset_mappings[j][segment_stop_index-1][0] > 0 or \
                                subtoken_offset_mappings[j][segment_stop_index][0] > 0:
                            # if breaking sentence in the middle of a token, truncate so whole token is in next segment
                            segment_stop_index -= 1
                        segments_ids.append(self.pad_list([tokenizer.cls_token_id] +
                                                          subtoken_ids[j][:segment_stop_index] +
                                                          [tokenizer.sep_token_id], max_segment_len,
                                                          tokenizer.convert_tokens_to_ids(tokenizer.pad_token)))
                        segments_clusters.append(self.pad_list([None] + subtoken_cluster_ids[j][:segment_stop_index]
                                                               + [None], max_segment_len, None))
                        segments_attention_mask.append(
                            self.pad_list([1] * (segment_stop_index + 2), max_segment_len))
                        sentence_map += [j] * (segment_stop_index + 2)
                        if use_metadata and speaker_ids:
                            segments_speakers.append(self.pad_list([-1] + subtoken_speakers[j][:segment_stop_index] +
                                                                   [-1], max_segment_len))
                        # remove already added data
                        subtoken_ids[j] = subtoken_ids[j][segment_stop_index:]
                        subtoken_cluster_ids[j] = subtoken_cluster_ids[j][segment_stop_index:]
                        if use_metadata and speaker_ids:
                            subtoken_speakers[j] = subtoken_speakers[j][segment_stop_index:]
                    current_segment_ids = subtoken_ids[j]
                    current_segment_cluster_ids = subtoken_cluster_ids[j]
                    current_sentence_map = [j] * len(subtoken_ids[j])
                    if use_metadata and speaker_ids:
                        current_segment_speaker_ids = subtoken_speakers[j]
            # get last segment
            segments_ids.append(self.pad_list([tokenizer.cls_token_id] + current_segment_ids +
                                              [tokenizer.sep_token_id], max_segment_len,
                                              tokenizer.convert_tokens_to_ids(tokenizer.pad_token)))
            segments_clusters.append(self.pad_list([None] + current_segment_cluster_ids + [None],
                                                   max_segment_len, None))
            segments_attention_mask.append(self.pad_list([1] * (len(current_segment_ids) + 2), max_segment_len))
            sentence_map += [current_sentence_map[0]] + current_sentence_map + [current_sentence_map[-1]]

            if use_metadata:
                segments_speakers.append(self.pad_list([-1] + current_segment_speaker_ids + [-1],
                                                       max_segment_len))
            # create document level info (cluster indices, cluster ids, sentence map, genre)
            gold_starts = []
            gold_ends = []
            gold_cluster_ids = []
            current_offset = 0
            for j in range(len(segments_clusters)):
                # loop over segments and create gold start/ends, ids
                valid_sub_tokens = sum(segments_attention_mask[j])
                cluster_sightings = {}  # keys: clusterids, values: all indices where that clusterid is observed
                for k in range(valid_sub_tokens):
                    if segments_clusters[j][k]:
                        if isinstance(segments_clusters[j][k], tuple) or isinstance(segments_clusters[j][k], list):
                            for clus_id in segments_clusters[j][k]:
                                if clus_id in cluster_sightings:
                                    cluster_sightings[clus_id].append(k)
                                else:
                                    cluster_sightings[clus_id] = [k]
                        elif segments_clusters[j][k] >= 0:
                            if segments_clusters[j][k] in cluster_sightings:
                                cluster_sightings[segments_clusters[j][k]].append(k)
                            else:
                                cluster_sightings[segments_clusters[j][k]] = [k]
                for clus_id, indices in cluster_sightings.items():
                    indices_pointer = 0
                    while indices_pointer < len(indices):
                        gold_starts.append(indices[indices_pointer] + current_offset)
                        gold_cluster_ids.append(clus_id)
                        while indices_pointer < len(indices) - 1 and indices[indices_pointer] == \
                                indices[indices_pointer + 1] - 1:
                            indices_pointer += 1
                        gold_ends.append(indices[indices_pointer] + current_offset)
                        indices_pointer += 1
                current_offset += valid_sub_tokens
            # sort cluster data by cluster start
            cluster_data = sorted(zip(gold_starts, gold_ends, gold_cluster_ids), key=lambda x: x[0])
            cluster_data = [list(t) for t in zip(*cluster_data)]
            if len(cluster_data) == 0:
                cluster_data = [[], [], []]
            self.examples.append([torch.as_tensor(segments_ids), torch.as_tensor(segments_attention_mask),
                                  torch.as_tensor(sentence_map), torch.as_tensor(cluster_data[0]),
                                  torch.as_tensor(cluster_data[1]), torch.as_tensor(cluster_data[2])])
            if use_metadata:
                self.examples[-1].append(torch.as_tensor(segments_speakers))
                self.examples[-1].append(torch.as_tensor(genres[i] if isinstance(genres[i], int) else
                                                         genre_to_id[genres[i]]))

        # self.examples: list of tensors in following order:
        #   ids, attentionmask, sentencemap, goldstarts, goldends, clusterids, speaker ids, genre

    @staticmethod
    def pad_list(values, target_length, pad_value=0):
        """
        Function to pad a list of values to a specific length, appending the pad_value to the end of the list.

        Parameters
        ----------
        values: List
        target_length: int
        pad_value: value pad the list with

        Returns
        -------
        list of values, padded to target length
        """
        while len(values) < target_length:
            values.append(pad_value)
        return values

    @staticmethod
    def get_subtoken_data(token_data, offset_mapping):
        """
        Function to map token data to sub tokens. For example, if a token is split into two sub-tokens,
        the cluster id (or speaker id) for the token needs to be associated with both sub-tokens.

        Parameters
        ----------
        token_data: cluster ids of tokens
        offset_mapping: for each sub-token, a (start index, end index) tuple of indices into it's original token
            As returned by a transformers.tokenizer if return_offsets_mapping=True.

        Returns
        -------
        List containing cluster ids for each token
        """
        token_index = -1
        sub_token_data = []
        for (start, _) in offset_mapping:
            if start == 0:
                token_index += 1
            sub_token_data.append(token_data[token_index])
        return sub_token_data

    def __len__(self):
        return len(self.examples)

    def __getitem__(self, i):
        """
        Each example is 1 document that consists of the following:
        input_ids: tensor of size (num_segments, sequence_length)
            input token ids
        attention_mask: tensor of size (num_segments, sequence_length)
            attention mask of input segments
        sentence_map: tensor of size (num_tokens)
            sentence id for each input token in input document
        speaker_ids: tensor of size (num_segments, sequence_length)
            speaker ids for each token (only used if self.use_metadata is True)
        genre: tensor of size (1)
            genre id for document
        gold_starts: tensor of size (labeled)
            start token indices (in flattened document) of labeled spans
        gold_ends: tensor of size (labeled)
            end token indices (in flattened document) of labeled spans
        cluster_ids: tensor of size (labeled)
            cluster ids of each labeled span
        """
        if self.use_metadata:
            ids, attention_mask, sentence_map, gold_starts, gold_ends, cluster_ids, speaker_ids, genre = \
                self.examples[i]
        else:
            ids, attention_mask, sentence_map, gold_starts, gold_ends, cluster_ids = self.examples[i]
        if len(ids) > self.max_segments:
            sentence_offset = random.randint(0, len(ids) - self.max_segments)
            token_offset = attention_mask[:sentence_offset].sum()
            ids = ids[sentence_offset:sentence_offset + self.max_segments]
            attention_mask = attention_mask[sentence_offset:sentence_offset + self.max_segments]
            num_tokens = attention_mask.sum()
            sentence_map = sentence_map[token_offset:token_offset + num_tokens]
            gold_spans = torch.logical_and(gold_ends >= token_offset, gold_starts < token_offset + num_tokens)
            gold_starts = gold_starts[gold_spans] - token_offset
            gold_ends = gold_ends[gold_spans] - token_offset
            cluster_ids = cluster_ids[gold_spans]
            if self.use_metadata:
                speaker_ids = speaker_ids[sentence_offset:sentence_offset + self.max_segments]
        inputs = {
            "input_ids": ids,
            "attention_mask": attention_mask,
            "sentence_map": sentence_map,
            "gold_starts": gold_starts,
            "gold_ends": gold_ends,
            "cluster_ids": cluster_ids
        }
        if self.use_metadata:
            inputs.update({"speaker_ids": speaker_ids, "genre": genre})
        return inputs


class MangoesLineByLineDatasetForNSP(transformers.TextDatasetForNextSentencePrediction):
    """
    Subclass of Huggingface TextDatasetForNextSentencePrediction that supports multiple input files.
    Used for next sentence prediction task.

    Input file format:

    (1) One sentence per line. These should ideally be actual sentences, not entire paragraphs or arbitrary spans of
    text.
    (2) Blank lines between documents. Document boundaries are needed so that the "next sentence prediction" task
    doesn't span between documents. Opening a new file will assume a new document as well.
    (3) Assumes different input files contain different documents. (ie, a document cannot span multiple files)


    Example:
    I am very happy. (new line) Here is the second sentence. (new line) (new line) A new document.

    Parameters
    ---------
    filenames: str or List[str]
        paths to files to include in dataset.
    tokenizer: transformers.Tokenizer
    short_seq_probability: float
        probability to sample shorter sequences. Used in the original BERT paper.
    nsp_probability: float
        probability to sample random next sentence for next sentence prediction task.
    max_len: int
        max length of input sequences, if None, will default to tokenizer.model_max_length
    encoding: str
        encoding of text files
    """
    def __init__(self, filenames, tokenizer, short_seq_probability=0.1, nsp_probability=0.5, max_len=None,
                 encoding=None):
        if isinstance(filenames, str):
            filenames = [filenames]
        max_len = max_len if max_len else tokenizer.model_max_length
        self.block_size = max_len - tokenizer.num_special_tokens_to_add(pair=True)
        self.short_seq_probability = short_seq_probability
        self.nsp_probability = nsp_probability
        self.encoding = encoding
        self.tokenizer = tokenizer
        self.documents = [[]]
        with fileinput.input(filenames, mode="r", openhook=None if not self.encoding else
                             fileinput.hook_encoded(self.encoding)) as f:
            while True:
                line = f.readline()
                if not line:
                    break
                line = line.strip()
                # Empty lines are used as document delimiters, or opening new file
                if (fileinput.isfirstline() or not line) and len(self.documents[-1]) != 0:
                    self.documents.append([])
                tokens = tokenizer.tokenize(line)
                tokens = tokenizer.convert_tokens_to_ids(tokens)
                if tokens:
                    self.documents[-1].append(tokens)
        self.examples = []
        for doc_index, document in enumerate(self.documents):
            self.create_examples_from_document(document, doc_index)


class MangoesMultipleChoiceDataset(Dataset):
    """
    Subclass of Torch Dataset for multiple choice datasets such as SWAG. Currently meant to work with BERT models.

    For information on how multiple choice datasets are formatted using this class, see
    https://github.com/google-research/bert/issues/38

    And this link for explanation of Huggingface's multiple choice models:
    https://github.com/huggingface/transformers/issues/7701#issuecomment-707149546

    Parameters
    ----------
    tokenizer: transformers.BertTokenizerFast
        tokenizer to use
    question_texts: List of str
        The texts corresponding to the questions/contexts.
    choices_texts: List of str
        The texts corresponding to the answer choices
    labels: List of int
        The indices of the correct answers
    max_seq_length:int
        The maximum total input sequence length after tokenization. if None, will default to tokenizer.model_max_length.
    """
    def __init__(self, tokenizer, question_texts, choices_texts, labels, max_seq_length=None):
        self.tokenizer = tokenizer
        self.question_texts = question_texts
        self.choices_texts = choices_texts
        self.labels = labels
        self.max_seq_len = max_seq_length

    def __len__(self):
        return len(self.labels)

    def __getitem__(self, i):
        num_choices = len(self.choices_texts[i])
        questions = [self.question_texts[i]] * num_choices

        inputs = self.tokenizer(questions, self.choices_texts[i], padding="max_length", max_length=self.max_seq_len,
                                truncation=True, return_tensors="pt")
        inputs["labels"] = torch.as_tensor(int(self.labels[i]))
        return inputs


class MultipleLearnRateFineTuneTrainer(transformers.Trainer):
    """
    Subclass of Huggingface Trainer to accept different learning rates for base model parameters and task specific
    parameters, in the context of a fine-tuning task.

    Parameters
    ---------
    task_learn_rate: float
        Learning rate to be used for task specific parameters, (base parameters will use the normal, ie already defined
        in args, learn rate)
    base_keyword: str
        String to be used to differentiate base model and task specific parameters. All named parameters that have
        "base_keyword" somewhere in the name will be considered part of the base model, while all parameters that don't
        will be considered part of the task specific parameters.
    For documentation of the rest of the init parameters, see
        https://huggingface.co/transformers/main_classes/trainer.html#id1
    """
    def __init__(
        self,
        task_learn_rate,
        base_keyword="bert",
        model=None,
        args=None,
        data_collator=None,
        train_dataset=None,
        eval_dataset=None,
        tokenizer=None,
        model_init=None,
        compute_metrics=None,
        callbacks=None,
        optimizers=(None, None)
    ):
        self.task_learn_rate = task_learn_rate
        self.base_keyword = base_keyword
        super(MultipleLearnRateFineTuneTrainer, self).__init__(model, args, data_collator, train_dataset, eval_dataset,
                                                               tokenizer, model_init, compute_metrics, callbacks,
                                                               optimizers)

    def create_optimizer_and_scheduler(self, num_training_steps):
        """
        Setup the optimizer and the learning rate scheduler.

        This will use AdamW. If you want to use something else (ie, a different optimizer and multiple learn rates), you
        can subclass and override this method in a subclass.
        """
        if self.optimizer is None:
            no_decay = ["bias", "LayerNorm.weight"]
            optimizer_grouped_parameters = [
                {
                    "params": [p for n, p in self.model.named_parameters() if not any(nd in n for nd in no_decay) and
                               self.base_keyword in n],
                    "weight_decay": self.args.weight_decay,
                },
                {
                    "params": [p for n, p in self.model.named_parameters() if any(nd in n for nd in no_decay) and
                               self.base_keyword in n],
                    "weight_decay": 0.0,
                },
                {
                    "params": [p for n, p in self.model.named_parameters() if not any(nd in n for nd in no_decay) and
                               self.base_keyword not in n],
                    "weight_decay": self.args.weight_decay,
                    "lr": self.task_learn_rate,
                },
                {
                    "params": [p for n, p in self.model.named_parameters() if any(nd in n for nd in no_decay) and
                               self.base_keyword not in n],
                    "weight_decay": 0.0,
                    "lr": self.task_learn_rate,
                },
            ]
            if self.args.adafactor:
                optimizer_cls = transformers.Adafactor
                optimizer_kwargs = {"scale_parameter": False, "relative_step": False}
            else:
                optimizer_cls = transformers.AdamW
                optimizer_kwargs = {
                    "betas": (self.args.adam_beta1, self.args.adam_beta2),
                    "eps": self.args.adam_epsilon,
                }
            optimizer_kwargs["lr"] = self.args.learning_rate
            self.optimizer = optimizer_cls(optimizer_grouped_parameters, **optimizer_kwargs)

        if self.lr_scheduler is None:
            self.lr_scheduler = transformers.get_scheduler(
                self.args.lr_scheduler_type,
                self.optimizer,
                num_warmup_steps=self.args.warmup_steps,
                num_training_steps=num_training_steps,
            )


class CoreferenceFineTuneTrainer(MultipleLearnRateFineTuneTrainer):
    """
    Subclass of the Mangoes MultipleLearnRateFineTuneTrainer that does not collate examples into batches, for use in the
    default Coreference Fine Tuning Trainer.

    This method introduces a dummy batch collation method, because the batches in the implemented fine tuning method
    (see paper below) are exactly 1 document each, and are pre-collated in the dataset class.
    This is based on the independent variant of the coreference resolution method described in
    https://arxiv.org/pdf/1908.09091.pdf.

    For documentation of the init parameters, see the documentation for MangoesMultipleLearnRateFineTuneTrainer
    """
    def __init__(
            self,
            task_learn_rate,
            base_keyword="bert",
            model=None,
            args=None,
            train_dataset=None,
            eval_dataset=None,
            tokenizer=None,
            model_init=None,
            compute_metrics=None,
            callbacks=None,
            optimizers=(None, None)
    ):
        def collate_fn(batch):
            return batch[0]
        super(CoreferenceFineTuneTrainer, self).__init__(task_learn_rate, base_keyword, model, args,
                                                         collate_fn, train_dataset, eval_dataset, tokenizer,
                                                         model_init, compute_metrics, callbacks, optimizers)


class IterableCompatibleTrainer(transformers.Trainer):
    """
    Subclass of Huggingface Trainer to accept Iterable datasets
    """

    def get_train_dataloader(self):
        """
        Returns
        -------
        torch DataLoader
        """
        if self.train_dataset is None:
            raise ValueError("Trainer: training requires a train_dataset.")
        if not isinstance(self.train_dataset, IterableDataset):
            train_sampler = self._get_train_sampler()
        else:
            train_sampler = None
        return DataLoader(
            self.train_dataset,
            batch_size=self.args.train_batch_size,
            sampler=train_sampler,
            collate_fn=self.data_collator,
            drop_last=self.args.dataloader_drop_last,
            num_workers=self.args.dataloader_num_workers,
        )

    def get_eval_dataloader(self, eval_dataset=None):
        """
        Parameters
        ----------
        eval_dataset: torch Dataset or IterableDataset

        Returns
        -------
        torch DataLoader
        """
        if eval_dataset is None and self.eval_dataset is None:
            raise ValueError("Trainer: evaluation requires an eval_dataset.")
        if eval_dataset is not None and not isinstance(eval_dataset, collections.abc.Sized):
            raise ValueError("eval_dataset must implement __len__")
        eval_dataset = eval_dataset if eval_dataset is not None else self.eval_dataset
        if not isinstance(eval_dataset, IterableDataset):
            eval_sampler = self._get_eval_sampler(eval_dataset)
        else:
            eval_sampler = None
        return DataLoader(
            eval_dataset,
            sampler=eval_sampler,
            batch_size=self.args.eval_batch_size,
            collate_fn=self.data_collator,
            drop_last=self.args.dataloader_drop_last,
            num_workers=self.args.dataloader_num_workers,
        )
